/*
 *  arch-tag: Implementation of several functions to manage the covers
 *
 *  Copyright (C) 2004,2005 Marc Pavot <m.pavot@laposte.net>
 *  Copyright (C) 2005 Christophe Fergeau <teuf@gnome.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <gtk/gtk.h>
#include <libgnomevfs/gnome-vfs.h>
#include <libgnome/gnome-i18n.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <string.h>

#include "rb-cover.h"
#include "rb-file-helpers.h"
#include "rb-debug.h"
#include "rb-preferences.h"
#include "../widgets/eel-gconf-extensions.h"

#define READ_CHUNK_SIZE 8192

#define COVER_SMALL "ImageUrlSmall"
#define COVER_MEDIUM "ImageUrlMedium"
#define COVER_LARGE "ImageUrlLarge"

static void read_file_read_chunk (RBCoverReadFileHandle *handle);
static void rb_cover_save_cover_open_cb (GnomeVFSAsyncHandle *handle,
					 GnomeVFSResult result,
					 gpointer callback_data);
static const char *valid_cover_names[] = {
	"folder",
	"cover",
	NULL
};

static const char *valid_cover_extensions[] = {
	"jpg",
	"jpeg",
	"png",
	"gif",
	NULL
};

GnomeVFSResult
rb_cover_read_entire_file (const char *uri,
			   int *file_size,
			   char **file_contents)
{
	GnomeVFSResult result;
	GnomeVFSHandle *handle;
	char *buffer;
	GnomeVFSFileSize total_bytes_read;
	GnomeVFSFileSize bytes_read;

	*file_size = 0;
	*file_contents = NULL;
	/* Open the file. */
	result = gnome_vfs_open (&handle, uri, GNOME_VFS_OPEN_READ);

	if (result != GNOME_VFS_OK) {
		return result;
	}

	/* Read the whole thing. */
	buffer = NULL;
	total_bytes_read = 0;
	do {
		buffer = g_realloc (buffer, total_bytes_read + READ_CHUNK_SIZE);
		result = gnome_vfs_read (handle,
				buffer + total_bytes_read,
				READ_CHUNK_SIZE,
				&bytes_read);
		if (result != GNOME_VFS_OK && result != GNOME_VFS_ERROR_EOF) {
			g_free (buffer);
			gnome_vfs_close (handle);
			return result;
		}

		/* Check for overflow. */
		if (total_bytes_read + bytes_read < total_bytes_read) {
			g_free (buffer);
			gnome_vfs_close (handle);
			return GNOME_VFS_ERROR_TOO_BIG;
		}

		total_bytes_read += bytes_read;
	} while (result == GNOME_VFS_OK);
	/* Close the file. */
	result = gnome_vfs_close (handle);
	if (result != GNOME_VFS_OK) {
		g_free (buffer);
		return result;
	}

	/* Return the file. */
	*file_size = total_bytes_read;
	*file_contents = g_realloc (buffer, total_bytes_read);

	return GNOME_VFS_OK;
}

gchar *
rb_cover_make_cover_path (const gchar *artist,
			  const gchar *album,
			  RBCoverHomeCoversSize cover_size)
{	
	char *cover_path, *tmp;
	char *filename;
	const char *to_strip = "#/";

	if (cover_size == SMALL_COVER)
		filename = g_strdup_printf ("SMALL%s-%s.jpg", artist, album);
	else
		filename = g_strdup_printf ("%s-%s.jpg", artist, album);

	for (tmp = filename; *tmp != '\0'; tmp++) {
		if (strchr (to_strip, *tmp))
			*tmp = ' ';
	}	
	
	cover_path = g_build_filename (rb_dot_dir (), "covers", filename, NULL);

	g_free (filename);
	return cover_path;
}

gboolean
rb_cover_cover_exists (const gchar *artist,
		       const gchar *album)
{
	gchar *cover_path, *small_cover_path;
	gboolean result;

	cover_path = rb_cover_make_cover_path (artist,
					       album,
					       NORMAL_COVER);

	small_cover_path = rb_cover_make_cover_path (artist,
						     album,
						     SMALL_COVER);

	result = (rb_uri_exists (cover_path) && rb_uri_exists (small_cover_path));
	
	g_free (cover_path);
	g_free (small_cover_path);

	return result;
}

static gboolean
rb_album_cover_is_cover_file_name (const char *filename)
{
	int i, j;
	char *valid_filename;

	for (i = 0; valid_cover_names[i] != NULL; i++) {
		for (j = 0; valid_cover_extensions[j] != NULL; j++) {
			valid_filename = g_strconcat (valid_cover_names[i], ".", valid_cover_extensions[j], NULL);
			if (g_ascii_strcasecmp (filename, valid_filename) == 0) {
				return TRUE;
			}
			g_free (valid_filename);
		}
	}
	return FALSE;
}

gchar *
rb_cover_find_local_uri (const char *location)
{
	char *path;
	char *cover_uri = NULL;
	GnomeVFSURI *uri;
	GList *fileinfos;
	GList *it;
	GnomeVFSResult res;

	if (rb_uri_is_iradio(location))
		return NULL;

	uri = gnome_vfs_uri_new (location);
	if (uri == NULL) {
		return NULL;
	}
	path = gnome_vfs_uri_extract_dirname (uri);
	gnome_vfs_uri_unref (uri);
	if (path == NULL) {
		return NULL;
	}

	res = gnome_vfs_directory_list_load (&fileinfos, path, 
					     GNOME_VFS_FILE_INFO_DEFAULT);
	if (res != GNOME_VFS_OK) {
		return NULL;
	}
	
	cover_uri = NULL;
	for (it = fileinfos; it != NULL; it = it->next) {
		GnomeVFSFileInfo *info;
		info = (GnomeVFSFileInfo*)it->data;
		if (rb_album_cover_is_cover_file_name (info->name) && rb_cover_size_is_valid (info->size)) {
			cover_uri = g_strconcat (path, "/", info->name, NULL);
			break;
		}
	}

	g_list_foreach (fileinfos, (GFunc) gnome_vfs_file_info_unref, NULL);
	g_list_free (fileinfos);
	g_free (path);
	return cover_uri;
}

GnomeVFSResult
rb_cover_load_local_cover (const char *location,
			   int *file_size,
			   char **file_contents)
{
	char *cover_uri;
	
	cover_uri = rb_cover_find_local_uri (location);

	if (!cover_uri)
		return GNOME_VFS_ERROR_BAD_PARAMETERS;

	return rb_cover_read_entire_file (cover_uri,
					  file_size,
					  file_contents);
}

GList * 
rb_cover_parse_xml_file (char *xmldata,
			 int size,
			 RBCoverAmazonOperation operation,
			 RBCoverAmazonCoversSize cover_size)
{
	xmlDocPtr doc;
	xmlNodePtr cur;
	xmlChar *key;
	GList *cover_uris = NULL;
	char *cover_size_text;

	switch (cover_size)
	{
	case AMAZON_SMALL_COVER:
		cover_size_text = COVER_SMALL;
		break;
	case AMAZON_MEDIUM_COVER:
		cover_size_text = COVER_MEDIUM;
		break;
	case AMAZON_LARGE_COVER:
		cover_size_text = COVER_LARGE;
		break;
	default:
		cover_size_text = COVER_MEDIUM;
		break;
	}

	rb_debug ("Parsing XML file from Amazon");

	doc = xmlParseMemory (xmldata, size);

	if (doc == NULL ) {
		return cover_uris;
	}

	cur = xmlDocGetRootElement(doc);
	
	if (cur == NULL) {
		xmlFreeDoc(doc);
		return cover_uris;
	}

	if (xmlStrcmp(cur->name, (const xmlChar *) "ProductInfo")) {
		xmlFreeDoc(doc);
		return cover_uris;
	}
	cur = cur->xmlChildrenNode;
	while (cur != NULL) {
		if (!xmlStrcmp (cur->name, (const xmlChar *)"Details")){
			xmlNodePtr cur2;
			cur2 = cur->xmlChildrenNode;
			while (cur2 != NULL) {
				if ((!xmlStrcmp (cur2->name, (const xmlChar *) cover_size_text))) {
					key = xmlNodeListGetString (doc, cur2->xmlChildrenNode, 1);
					cover_uris = g_list_append (cover_uris, key);
					if (operation == GET_FIRST_COVER) {
						xmlFreeDoc(doc);
						return cover_uris;
					}
				}
				cur2 = cur2->next;
			}
		}
		cur = cur->next;
	}

	xmlFreeDoc(doc);

	return cover_uris;
}

static char *strnrepl (char *s, int size, const char *old, const char *new)
{
	char *ptr = s;
	int left = strlen (s);
	int avail = size - (left+1);
	int oldlen = strlen (old);
	int newlen = strlen (new);
	int diff = newlen - oldlen;

	while (left >= oldlen) {
		if (strncmp (ptr, old, oldlen) != 0) {
			left--;
			ptr++;
			continue;
		}
		if (diff > avail)
			break;
		if (diff != 0)
			memmove (ptr+newlen, ptr+oldlen, left);
		strncpy (ptr, new, newlen);
		ptr += newlen;
		left -= oldlen;
	}
	return s;
}

char *
rb_cover_make_xml_uri (const char *artist, 
		       const char *album)
{
	char *xml_uri;
	char *keywords;
	char *locale;
	char *music_mode;
	char *ext;
	char *tmp;
	int i;

	const char *mykey = "1BDCAEREYT743R9SXE02";
	const gchar *to_replace[] = {"é", "è", "ê", "à", "ù", "ç", "#", "/", "?", "'", "-", "\"", NULL};
	const gchar *replace_to[] = {"e", "e", "e", "a", "u", "c", "" , "" , "" , " ", " ", " ", NULL};
	const char *unknown = _("Unknown");

	if (!album || !artist)
		return NULL;

	if (!strcmp (artist, unknown))
		return NULL;

	if (!strcmp (album, unknown))
		keywords = g_strdup (artist);
	else
		keywords = g_strdup_printf ("%s %s", artist, album);

	for (i = 0; to_replace[i]; i++) {
		if (replace_to[i])
			keywords = strnrepl (keywords, strlen (keywords) + 1, to_replace[i], replace_to[i]); 
	}

	tmp = gnome_vfs_escape_string (keywords);
	g_free (keywords);
	keywords = tmp;

	locale = eel_gconf_get_string (CONF_COVER_AMAZON_COUNTRY);

	if (!locale)
		locale = "com";

	music_mode = "music";
	ext = "com";

	if (!strcmp (locale, "fr")) {
		g_free (music_mode);
		music_mode = "music-fr";
	}

	if (!strcmp (locale, "jp")) {
		g_free (music_mode);
		music_mode = "music-jp";
		g_free (ext);
		ext = "co.jp";
	}

	if (!strcmp (locale, "com")) {
		g_free (locale);
		locale = "us";
	}

	xml_uri = g_strdup_printf ("http://xml.amazon.%s/onca/xml3?t=webservices-20&dev-t=%s&KeywordSearch=%s&mode=%s&locale=%s&type=lite&page=1&f=xml", ext, mykey, keywords, music_mode, locale);

	g_free (keywords);

	return xml_uri;
}

GnomeVFSResult
rb_cover_load_amazon_covers (const char *artist,
			     const char *album,
			     GArray **file_size,
			     GList **file_contents,
			     RBCoverAmazonOperation operation,
			     RBCoverAmazonCoversSize cover_size)
{
	char *xml_uri;
	int xml_size;
	char *xml_data;
	GList *cover_uris, *temp;
	GnomeVFSResult result, result_temp;
	int temp_size;
	char *temp_contents;

	xml_uri = rb_cover_make_xml_uri (artist, 
					 album);

	if (!xml_uri)
		return GNOME_VFS_ERROR_BAD_PARAMETERS;

	result = rb_cover_read_entire_file (xml_uri,
					    &xml_size,
					    &xml_data);
	g_free (xml_uri);

	if (result != GNOME_VFS_OK) {
		g_free (xml_data);
		return result;
	}

	cover_uris = rb_cover_parse_xml_file (xml_data,
					      xml_size,
					      operation,
					      cover_size);

	g_free (xml_data);

	result = GNOME_VFS_ERROR_NOT_FOUND;

	for (temp = cover_uris; temp; temp = temp->next) {
		if (temp->data) {
			result_temp = rb_cover_read_entire_file (temp->data,
								 &temp_size,
								 &temp_contents);
			if (result_temp == GNOME_VFS_OK && rb_cover_size_is_valid (temp_size)) {
				result = result_temp;
				g_array_append_val (*file_size, temp_size);
				*file_contents = g_list_append (*file_contents, temp_contents);
			}
		}
	}
	g_list_foreach (cover_uris, (GFunc) g_free, NULL);
	g_list_free (cover_uris);

	return result;
}

void
rb_cover_remove_cover (const gchar *artist,
		       const gchar *album)
{
	gchar *small_cover_path;
	gchar *cover_path;

	/* delete the small cover*/
	small_cover_path = rb_cover_make_cover_path (artist, album, SMALL_COVER);
	if (rb_uri_exists (small_cover_path))
		gnome_vfs_unlink (small_cover_path);
	g_free (small_cover_path);

	/* delete the normal cover*/
	cover_path = rb_cover_make_cover_path (artist, album, NORMAL_COVER);
	if (rb_uri_exists (cover_path))
		gnome_vfs_unlink (cover_path);
	g_free (cover_path);
}

struct RBCoverReadFileHandle {
	GnomeVFSAsyncHandle *handle;
	RBCoverReadFileCallback callback;
	gpointer callback_data;
	gboolean is_open;
	char *buffer;
	GnomeVFSFileSize bytes_read;
};

/* When close is complete, there's no more work to do. */
static void
read_file_close_callback (GnomeVFSAsyncHandle *handle,
			  GnomeVFSResult result,
			  gpointer callback_data)
{
}

/* Do a close if it's needed.
 * Be sure to get this right, or we have extra threads hanging around.
 */
static void
read_file_close (RBCoverReadFileHandle *read_handle)
{
	if (read_handle->is_open) {
		gnome_vfs_async_close (read_handle->handle,
				       read_file_close_callback,
				       NULL);
		read_handle->is_open = FALSE;
	}
}

/* Close the file and then tell the caller we succeeded, handing off
 * the buffer to the caller.
 */
static void
read_file_succeeded (RBCoverReadFileHandle *read_handle)
{
	read_file_close (read_handle);
	
	/* Reallocate the buffer to the exact size since it might be
	 * around for a while.
	 */
	(* read_handle->callback) (GNOME_VFS_OK,
				   read_handle->bytes_read,
				   g_realloc (read_handle->buffer,
					      read_handle->bytes_read),
				   read_handle->callback_data);

	g_free (read_handle);
}

/* Tell the caller we failed. */
static void
read_file_failed (RBCoverReadFileHandle *read_handle, GnomeVFSResult result)
{
	read_file_close (read_handle);
	g_free (read_handle->buffer);
	
	(* read_handle->callback) (result, 0, NULL, read_handle->callback_data);
	g_free (read_handle);
}

/* A read is complete, so we might or might not be done. */
static void
read_file_read_callback (GnomeVFSAsyncHandle *handle,
			 GnomeVFSResult result,
			 gpointer buffer,
			 GnomeVFSFileSize bytes_requested,
			 GnomeVFSFileSize bytes_read,
			 gpointer callback_data)
{
	RBCoverReadFileHandle *read_handle;
	gboolean read_more;

	/* Do a few reality checks. */
	g_assert (bytes_requested == READ_CHUNK_SIZE);
	read_handle = callback_data;
	g_assert (read_handle->handle == handle);
	g_assert (read_handle->buffer + read_handle->bytes_read == buffer);
	g_assert (bytes_read <= bytes_requested);

	/* Check for a failure. */
	if (result != GNOME_VFS_OK && result != GNOME_VFS_ERROR_EOF) {
		read_file_failed (read_handle, result);
		return;
	}

	/* Check for the extremely unlikely case where the file size overflows. */
	if (read_handle->bytes_read + bytes_read < read_handle->bytes_read) {
		read_file_failed (read_handle, GNOME_VFS_ERROR_TOO_BIG);
		return;
	}

	/* Bump the size. */
	read_handle->bytes_read += bytes_read;

	/* Read more unless we are at the end of the file. */
	if (bytes_read == 0 || result != GNOME_VFS_OK)
		read_more = FALSE;
	else
		read_more = TRUE;

	if (read_more) {
		read_file_read_chunk (read_handle);
		return;
	}

	/* If at the end of the file, we win! */
	read_file_succeeded (read_handle);
}

/* Start reading a chunk. */
static void
read_file_read_chunk (RBCoverReadFileHandle *handle)
{
	handle->buffer = g_realloc (handle->buffer, handle->bytes_read + READ_CHUNK_SIZE);
	gnome_vfs_async_read (handle->handle,
			      handle->buffer + handle->bytes_read,
			      READ_CHUNK_SIZE,
			      read_file_read_callback,
			      handle);
}

/* Once the open is finished, read a first chunk. */
static void
read_file_open_callback (GnomeVFSAsyncHandle *handle,
			 GnomeVFSResult result,
			 gpointer callback_data)
{
	RBCoverReadFileHandle *read_handle;
	
	read_handle = callback_data;
	g_assert (read_handle->handle == handle);

	/* Handle the failure case. */
	if (result != GNOME_VFS_OK) {
		read_file_failed (read_handle, result);
		return;
	}

	/* Handle success by reading the first chunk. */
	read_handle->is_open = TRUE;

	read_file_read_chunk (read_handle);
}
/* Set up the read handle and start reading. */
static RBCoverReadFileHandle *
eel_read_file_async (const char *uri,
		     int priority,
		     RBCoverReadFileCallback callback,
		     gpointer callback_data)
{
	RBCoverReadFileHandle *handle;

	handle = g_new0 (RBCoverReadFileHandle, 1);

	handle->callback = callback;
	handle->callback_data = callback_data;

	gnome_vfs_async_open (&handle->handle,
			      uri,
			      GNOME_VFS_OPEN_READ,
			      priority,
			      read_file_open_callback,
			      handle);
	return handle;
}

/* Set up the read handle and start reading. */
RBCoverReadFileHandle *
rb_cover_read_entire_file_async (const char *uri,
			  	 int priority,
			  	 RBCoverReadFileCallback callback,
			  	 gpointer callback_data)
{
	return eel_read_file_async (uri, priority, callback, callback_data);
}

gboolean
rb_cover_size_is_valid (int size)
{
	return (size < 1024 * 200 && size > 900);
}

GnomeVFSResult
rb_cover_create_cover_dir (void)
{
	gchar *cover_dir;
	GnomeVFSResult result;

	cover_dir = g_build_filename (rb_dot_dir (), "covers", NULL);

	if (!rb_uri_exists (cover_dir)) 
		 result = gnome_vfs_make_directory (cover_dir, GNOME_VFS_PERM_USER_ALL
							       | GNOME_VFS_PERM_GROUP_READ
							       | GNOME_VFS_PERM_OTHER_READ);
	else
		result = GNOME_VFS_OK;

	g_free (cover_dir);
	return result;
}

static gboolean
rb_cover_can_overwrite_cover (const gchar *artist,
			      const gchar *album,
			      RBCoverOverwriteMode overwrite_mode)
{
	GtkWidget *dialog;
	gint retval;

	if (rb_cover_cover_exists (artist, album)) {
		switch (overwrite_mode) {
		case OVERWRITE_MODE_ASK:
			dialog = gtk_message_dialog_new (NULL,
							 GTK_DIALOG_MODAL,
							 GTK_MESSAGE_QUESTION,
							 GTK_BUTTONS_YES_NO,
							 "The cover already exists. Do you want to replace it?");

			retval = gtk_dialog_run (GTK_DIALOG (dialog));
			gtk_widget_destroy (dialog);
			if (retval != GTK_RESPONSE_YES)
				return FALSE;
			break;
		case OVERWRITE_MODE_REPLACE:
			break;
		case OVERWRITE_MODE_SKIP:
			return FALSE;
		default:
			return FALSE;
		}
	}

	return TRUE;
}

GnomeVFSResult
rb_cover_save_cover (const gchar *artist,
		     const gchar *album,
		     char *data,
		     int size,
		     RBCoverOverwriteMode overwrite_mode)
{
	GnomeVFSResult result;
	gchar *cover_path, *small_cover_path;
	GdkPixbufLoader *loader;
	GdkPixbuf *pixbuf, *small_pixbuf;

	if (!artist || !album || !data)
		return GNOME_VFS_ERROR_BAD_PARAMETERS;

	result = rb_cover_create_cover_dir ();
	if (result != GNOME_VFS_OK)
		return result;

	if (!rb_cover_can_overwrite_cover (artist, album, overwrite_mode))
		return GNOME_VFS_OK;

	cover_path = rb_cover_make_cover_path (artist,
					       album,
					       NORMAL_COVER);

	small_cover_path = rb_cover_make_cover_path (artist,
						     album,
						     SMALL_COVER);

	loader = gdk_pixbuf_loader_new ();
	result = GNOME_VFS_ERROR_BAD_PARAMETERS;

	if (gdk_pixbuf_loader_write (loader,
				     (const guchar *)data,
				     size,
				     NULL)) {
		gdk_pixbuf_loader_close (loader, NULL);

		pixbuf = gdk_pixbuf_loader_get_pixbuf (loader);
		small_pixbuf = gdk_pixbuf_scale_simple (pixbuf,
							TREE_COVER_SIZE,
							TREE_COVER_SIZE,
							GDK_INTERP_BILINEAR);

		if (gdk_pixbuf_save (pixbuf, cover_path, "jpeg", NULL, NULL) &&
		    gdk_pixbuf_save (small_pixbuf, small_cover_path, "jpeg", NULL, NULL)) {
			result = GNOME_VFS_OK;
		}
		g_object_unref (G_OBJECT (pixbuf));
		g_object_unref (G_OBJECT (small_pixbuf));
	}
	
	g_free (cover_path);
	g_free (small_cover_path);

	return result;
}

struct RBCoverSaveFileHandle {
	GnomeVFSAsyncHandle *handle;
	RBCoverSaveFileCallback callback;
	gchar *data;
	gchar *small_cover_data;
	int size;
	gpointer callback_data;
	gboolean is_small_cover;
	gchar *small_cover_path;
};

static void
rb_cover_save_small_cover (RBCoverSaveFileHandle *write_handle)
{
	GdkPixbufLoader *loader;
	GdkPixbuf *pixbuf, *small_pixbuf;
	gboolean ret_val;

	loader = gdk_pixbuf_loader_new ();

	if (gdk_pixbuf_loader_write (loader,
				     (const guchar *)write_handle->data,
				     write_handle->size,
				     NULL)) {
		gdk_pixbuf_loader_close (loader, NULL);

		pixbuf = gdk_pixbuf_loader_get_pixbuf (loader);
		small_pixbuf = gdk_pixbuf_scale_simple (pixbuf,
							TREE_COVER_SIZE,
							TREE_COVER_SIZE,
							GDK_INTERP_BILINEAR);

		ret_val = gdk_pixbuf_save_to_buffer (small_pixbuf,
						     &write_handle->small_cover_data,
						     (gsize *)&write_handle->size,
						     "jpeg", NULL, NULL);

		g_object_unref (G_OBJECT (pixbuf));
		g_object_unref (G_OBJECT (small_pixbuf));

		if (!ret_val) {
			(* write_handle->callback) (GNOME_VFS_ERROR_BAD_PARAMETERS,
						    write_handle->callback_data);
			g_free (write_handle->small_cover_path);
			g_free (write_handle->small_cover_data);
			g_free (write_handle);
		}				

		gnome_vfs_async_create (&write_handle->handle,
					write_handle->small_cover_path,
					GNOME_VFS_OPEN_WRITE,
					FALSE,
					GNOME_VFS_PERM_USER_ALL
					| GNOME_VFS_PERM_GROUP_READ
					| GNOME_VFS_PERM_OTHER_READ,
					GNOME_VFS_PRIORITY_DEFAULT,
					rb_cover_save_cover_open_cb,
					write_handle);
	}
}

static void
rb_cover_save_cover_close_cb (GnomeVFSAsyncHandle *handle,
			      GnomeVFSResult result,
			      gpointer callback_data)
{
	RBCoverSaveFileHandle *write_handle;

	write_handle = callback_data;
	g_assert (write_handle->handle == handle);

	if (write_handle->is_small_cover) {
		(* write_handle->callback) (GNOME_VFS_OK,
					    write_handle->callback_data);
		g_free (write_handle->small_cover_data);
		g_free (write_handle->small_cover_path);
		g_free (write_handle);
	} else {
		write_handle->is_small_cover = TRUE;
		rb_cover_save_small_cover (write_handle);
	}
}

static void
rb_cover_save_cover_close (RBCoverSaveFileHandle *write_handle)
{
	gnome_vfs_async_close (write_handle->handle,
			       rb_cover_save_cover_close_cb,
			       write_handle);
}

static void
rb_cover_save_cover_succeeded (RBCoverSaveFileHandle *write_handle)
{
	rb_cover_save_cover_close (write_handle);
}

static void
rb_cover_save_cover_failed (RBCoverSaveFileHandle *write_handle,
			    GnomeVFSResult result)
{
	rb_cover_save_cover_close (write_handle);
	
	(* write_handle->callback) (result,
				    write_handle->callback_data);

	g_free (write_handle->small_cover_data);
	g_free (write_handle->small_cover_path);
	g_free (write_handle);
}

static void
rb_cover_save_cover_write_cb (GnomeVFSAsyncHandle *handle,
			      GnomeVFSResult result,
			      gpointer buffer,
			      GnomeVFSFileSize bytes_requested,
			      GnomeVFSFileSize bytes_written,
			      gpointer callback_data)
{
	RBCoverSaveFileHandle *write_handle;

	write_handle = callback_data;
	g_assert (write_handle->handle == handle);

	/* Check for a failure. */
	if (result != GNOME_VFS_OK) {
		rb_cover_save_cover_failed (write_handle, result);
		return;
	}

	rb_cover_save_cover_succeeded (write_handle);
}

static void
rb_cover_save_cover_open_cb (GnomeVFSAsyncHandle *handle,
			     GnomeVFSResult result,
			     gpointer callback_data)
{
	RBCoverSaveFileHandle *write_handle;
	gchar *data_to_save;
	
	write_handle = callback_data;
	g_assert (write_handle->handle == handle);

	/* Handle the failure case. */
	if (result != GNOME_VFS_OK) {
		rb_cover_save_cover_failed (write_handle, result);
		return;
	}

	if (write_handle->is_small_cover)
		data_to_save = write_handle->small_cover_data;
	else 
		data_to_save = write_handle->data;

	gnome_vfs_async_write (write_handle->handle,
			       data_to_save,
			       write_handle->size,
	                       (GnomeVFSAsyncWriteCallback) rb_cover_save_cover_write_cb, 
			       write_handle);
}

RBCoverSaveFileHandle*
rb_cover_save_cover_async (const gchar *artist,
			   const gchar *album,
			   char *data,
			   int size,
			   RBCoverOverwriteMode overwrite_mode,
			   RBCoverSaveFileCallback callback,
			   gpointer callback_data)
{
	RBCoverSaveFileHandle *write_handle;
	GnomeVFSResult result;
	gchar *cover_path, *small_cover_path;

	if (!artist || !album || !data) {
		(* callback) (GNOME_VFS_ERROR_BAD_PARAMETERS,
			      callback_data);
		return NULL;
	}

	result = rb_cover_create_cover_dir ();
	if (result != GNOME_VFS_OK) {
		(* callback) (GNOME_VFS_OK,
			      callback_data);
		return NULL;
	}

	if (!rb_cover_can_overwrite_cover (artist, album, overwrite_mode)) {
		(* callback) (GNOME_VFS_OK,
			      callback_data);
		return NULL;
	}

	cover_path = rb_cover_make_cover_path (artist,
					       album,
					       NORMAL_COVER);

	small_cover_path = rb_cover_make_cover_path (artist,
						     album,
						     SMALL_COVER);

	write_handle = g_new0 (RBCoverSaveFileHandle, 1);

	write_handle->callback = callback;
	write_handle->callback_data = callback_data;
	write_handle->data = data;
	write_handle->size = size;
	write_handle->is_small_cover = FALSE;
	write_handle->small_cover_path = small_cover_path;

	gnome_vfs_async_create (&write_handle->handle,
				cover_path,
				GNOME_VFS_OPEN_WRITE,
				FALSE,
				GNOME_VFS_PERM_USER_ALL
				| GNOME_VFS_PERM_GROUP_READ
				| GNOME_VFS_PERM_OTHER_READ,
				GNOME_VFS_PRIORITY_DEFAULT,
				rb_cover_save_cover_open_cb,
				write_handle);

	g_free (cover_path);
	return write_handle;
}
