/*
 *  arch-tag: Implementation of metadata reading using GStreamer
 *
 *  Copyright (C) 2003,2004 Colin Walters <walters@verbum.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <config.h>
#include <libgnome/gnome-i18n.h>
#include <glib.h>
#include <gst/gst.h>
#include <gst/gsttag.h>
#include <string.h>

#include "rb-metadata.h"
#include "rb-debug.h"
#include "rb-file-helpers.h"
G_DEFINE_TYPE(RBMetaData, rb_metadata, G_TYPE_OBJECT)

static void rb_metadata_finalize (GObject *object);

/*
 * Tagger setup consists of two phases.
 * The first creates whatever elements are needed to demux the stream
 * and connects signal handlers to complete the pipeline when the
 * stream demuxer creates source pads.  The pipeline is then iterated
 * until the demuxer is finished.  For some formats, no such iteration
 * is required.
 *
 * The second phase then adds whatever other elements are required
 * to connect the tagger to the sink.
 */

typedef struct _RBTaggerSetupData
{
	gboolean complete;
	GstPad *pad;
} RBTaggerSetupData;

typedef gboolean (*RBSetupTaggerElem) (GstBin *, GstElement *, RBTaggerSetupData *);
typedef gboolean (*RBCompleteTaggerElem) (GstBin *, RBTaggerSetupData *, GstElement **, GstElement **);

struct RBMetadataGstType
{
	char *mimetype;
	RBSetupTaggerElem setup_tagger_func;
	RBCompleteTaggerElem complete_tagger_func;
	char *human_name;
};


struct RBMetaDataPrivate
{
	char *uri;
  
	GHashTable *metadata;

	GstElement *pipeline;
	GstElement *sink;

	/* Array of RBMetadataGstType */
	GPtrArray *supported_types;

	char *type;
	gboolean handoff;
	gboolean eos;
	gboolean non_audio;
	GError *error;
};


static void
rb_metadata_class_init (RBMetaDataClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

	object_class->finalize = rb_metadata_finalize;
}

static void
_tagger_new_pad_cb (GstElement *element, GstPad *pad, RBTaggerSetupData *data)
{
	data->complete = TRUE;
	data->pad = pad;
}

static void
_tagger_no_more_pads_cb (GstElement *element, RBTaggerSetupData *data)
{
	data->pad = NULL;
	data->complete = TRUE;
}

static gboolean
_tagger_setup_demuxer (GstBin *bin, GstElement *source, RBTaggerSetupData *data, const char *demuxer)
{
	GstElement *demux;
	demux = gst_element_factory_make (demuxer, demuxer);
	if (!demux)
		return FALSE;

	gst_bin_add (bin, demux);
	gst_element_link (source, demux);

	g_signal_connect (G_OBJECT (demux), "new-pad", G_CALLBACK (_tagger_new_pad_cb), data);
	g_signal_connect (G_OBJECT (demux), "no-more-pads", G_CALLBACK (_tagger_no_more_pads_cb), data);
	return TRUE;
}

static gboolean
rb_setup_flac_tagger (GstBin *bin, GstElement *source, RBTaggerSetupData *data)
{
	GstElement *t;

	t = gst_element_factory_make ("flactag", "flactag");
	if (!t)
		return FALSE;

	gst_bin_add (bin, t);
	gst_element_link (source, t);

	data->complete = TRUE;
	data->pad = NULL;
	return TRUE;
}

static gboolean
rb_complete_flac_tagger (GstBin *bin, RBTaggerSetupData *data, GstElement **tagger, GstElement **link_from)
{
	GstElement *t = gst_bin_get_by_name (bin, "flactag");
	*tagger = t;
	*link_from = t;
	return TRUE;
}

static gboolean
rb_setup_id3_tagger (GstBin *bin, GstElement *source, RBTaggerSetupData *data)
{
	return _tagger_setup_demuxer (bin, source, data, "id3demuxbin");
}

static gboolean
rb_complete_id3_tagger (GstBin *bin, RBTaggerSetupData *data, GstElement **tagger, GstElement **link_from)
{
	GstElement *mux;
	GstPad *muxpad;

	if (data->pad == NULL)
		return FALSE;
	
	/* hmm, check caps on source pad? */
	mux = gst_element_factory_make ("id3mux", "id3mux");
	if (!mux)
		return FALSE;
	gst_bin_add (bin, mux);
	muxpad = gst_element_get_pad (mux, "sink");
	if (!gst_pad_link (data->pad, muxpad))
		return FALSE;

	*tagger = mux;
	*link_from = mux;
	return TRUE;
}

#if OGG_TAGGING_ACTUALLY_WORKS
/*
 * ogg tagging doesn't work, seemingly due to gstreamer problems.
 * i don't think anyone has tried to do this with gstreamer before,
 * so i'm not particularly surprised.
 */
static gboolean
rb_setup_ogg_tagger (GstBin *bin, GstElement *source, RBTaggerSetupData *data)
{
	return _tagger_setup_demuxer (bin, source, data, "oggdemux");
}

static gboolean
rb_complete_ogg_tagger (GstBin *bin, RBTaggerSetupData *data, GstElement **tagger, GstElement **link_from)
{
	GstElement *mux;
	GstElement *t;
	GstPad *t_pad;
	GstPad *mux_pad;
	GstCaps *caps;
	GstStructure *structure;
	const char *mimetype;
	const char *tagger_type;

	if (data->pad == NULL)
		return FALSE;

	caps = gst_pad_get_caps (data->pad);
	structure = gst_caps_get_structure (caps, 0);
	mimetype = gst_structure_get_name (structure);
	if (strcmp (mimetype, "audio/x-vorbis") == 0) {
		tagger_type = "vorbistag";
	} else if (strcmp (mimetype, "audio/x-flac") == 0) {
		tagger_type = "flactag";
	} else {
		/* unsupported ogg type.. speex? */
		return FALSE;
	}

	t = gst_element_factory_make (tagger_type, tagger_type);
	if (!t)
		return FALSE;
	gst_bin_add (bin, t);
	t_pad = gst_element_get_pad (t, "sink");
	gst_pad_link (data->pad, t_pad);

	mux = gst_element_factory_make ("oggmux", "oggmux");
	if (!mux)
		return FALSE;
	gst_bin_add (bin, mux);

	mux_pad = gst_element_get_request_pad (mux, "sink_%d");
	t_pad = gst_element_get_pad (t, "src");
	gst_pad_link (t_pad, mux_pad);
	
	*tagger = t;
	*link_from = mux;
	return TRUE;
}
#endif

static void
add_supported_type (RBMetaData *md,
		    const char *mime,
		    RBSetupTaggerElem setup_tagger_func,
		    RBCompleteTaggerElem complete_tagger_func,
		    const char *human_name)
{
	struct RBMetadataGstType *type = g_new0 (struct RBMetadataGstType, 1);
	type->mimetype = g_strdup (mime);
	type->setup_tagger_func = setup_tagger_func;
	type->complete_tagger_func = complete_tagger_func;
	type->human_name = g_strdup (human_name);
	g_ptr_array_add (md->priv->supported_types, type);
}

static void
rb_metadata_init (RBMetaData *md)
{
	md->priv = g_new0 (RBMetaDataPrivate, 1);

	md->priv->supported_types = g_ptr_array_new ();
	
	/* the list of supported types serves two purposes:
	 * - it knows how to construct elements for tag writing
	 * - it knows human-readable names for MIME types so we can say
	 *     "There is no plugin available to play WMV files"
	 *     rather than " .. play video/x-ms-asf files".
	 *
	 * only registering types we have plugins for defeats the second purpose.
	 */
	add_supported_type (md, "application/x-id3", rb_setup_id3_tagger, rb_complete_id3_tagger, "MP3");
	add_supported_type (md, "audio/mpeg", rb_setup_id3_tagger, rb_complete_id3_tagger, "MP3");
	add_supported_type (md, "application/ogg", NULL, NULL /* rb_setup_ogg_tagger, rb_complete_ogg_tagger */, "Ogg");
	add_supported_type (md, "application/x-ape", NULL, NULL, "MonkeysAudio");
	add_supported_type (md, "audio/x-m4a", NULL, NULL, "MPEG-4");
	add_supported_type (md, "audio/x-musepack", NULL, NULL, "Musepack");
	add_supported_type (md, "video/x-ms-asf", NULL, NULL, "WMA");
	add_supported_type (md, "audio/x-flac", rb_setup_flac_tagger, rb_complete_flac_tagger, "FLAC");
	add_supported_type (md, "audio/x-mod", NULL, NULL, "MOD");
}

static void
rb_metadata_finalize (GObject *object)
{
	int i;
	RBMetaData *md;

	md = RB_METADATA (object);

	for (i = 0; i < md->priv->supported_types->len; i++) {
		struct RBMetadataGstType *type = g_ptr_array_index (md->priv->supported_types, i);
		g_free (type->mimetype);
		g_free (type->human_name);
		g_free (type);
	}
	g_ptr_array_free (md->priv->supported_types, TRUE);

	if (md->priv->metadata)
		g_hash_table_destroy (md->priv->metadata);

	if (md->priv->pipeline)
		gst_object_unref (GST_OBJECT (md->priv->pipeline));

	g_free (md->priv->type);
	g_free (md->priv->uri);

	g_free (md->priv);

	G_OBJECT_CLASS (rb_metadata_parent_class)->finalize (object);
}

RBMetaData *
rb_metadata_new (void)
{
	return RB_METADATA (g_object_new (RB_TYPE_METADATA, NULL));
}

static void
free_gvalue (GValue *val)
{
	g_value_unset (val);
	g_free (val);
}

static int
rb_metadata_gst_tag_to_field (const char *tag)
{
	if (!strcmp (tag, GST_TAG_TITLE))
		return RB_METADATA_FIELD_TITLE;
	else if (!strcmp (tag, GST_TAG_ARTIST))
		return RB_METADATA_FIELD_ARTIST;
	else if (!strcmp (tag, GST_TAG_ALBUM))
		return RB_METADATA_FIELD_ALBUM;
	else if (!strcmp (tag, GST_TAG_GENRE))
		return RB_METADATA_FIELD_GENRE;
	else if (!strcmp (tag, GST_TAG_COMMENT))
		return RB_METADATA_FIELD_COMMENT;
	else if (!strcmp (tag, GST_TAG_TRACK_NUMBER))
		return RB_METADATA_FIELD_TRACK_NUMBER;
	else if (!strcmp (tag, GST_TAG_TRACK_COUNT)) 
		return RB_METADATA_FIELD_MAX_TRACK_NUMBER;
	else if (!strcmp (tag, GST_TAG_ALBUM_VOLUME_NUMBER))
		return RB_METADATA_FIELD_DISC_NUMBER;
	else if (!strcmp (tag, GST_TAG_ALBUM_VOLUME_COUNT))
		return RB_METADATA_FIELD_MAX_TRACK_NUMBER;
	else if (!strcmp (tag, GST_TAG_DESCRIPTION))
		return RB_METADATA_FIELD_DESCRIPTION;
	else if (!strcmp (tag, GST_TAG_DATE))
		return RB_METADATA_FIELD_RELEASE_DATE;
	else if (!strcmp (tag, GST_TAG_VERSION))
		return RB_METADATA_FIELD_VERSION;
	else if (!strcmp (tag, GST_TAG_ISRC))
		return RB_METADATA_FIELD_ISRC;
	else if (!strcmp (tag, GST_TAG_ORGANIZATION))
		return RB_METADATA_FIELD_ORGANIZATION;
	else if (!strcmp (tag, GST_TAG_COPYRIGHT))
		return RB_METADATA_FIELD_COPYRIGHT;
	else if (!strcmp (tag, GST_TAG_CONTACT))
		return RB_METADATA_FIELD_CONTACT;
	else if (!strcmp (tag, GST_TAG_LICENSE))
		return RB_METADATA_FIELD_LICENSE;
	else if (!strcmp (tag, GST_TAG_PERFORMER))
		return RB_METADATA_FIELD_PERFORMER;
	else if (!strcmp (tag, GST_TAG_DURATION))
		return RB_METADATA_FIELD_DURATION;
	else if (!strcmp (tag, GST_TAG_CODEC))
		return RB_METADATA_FIELD_CODEC;
	else if (!strcmp (tag, GST_TAG_BITRATE))
		return RB_METADATA_FIELD_BITRATE;
	else if (!strcmp (tag, GST_TAG_TRACK_GAIN))
		return RB_METADATA_FIELD_TRACK_GAIN;
	else if (!strcmp (tag, GST_TAG_TRACK_PEAK))
		return RB_METADATA_FIELD_TRACK_PEAK;
	else if (!strcmp (tag, GST_TAG_ALBUM_GAIN))
		return RB_METADATA_FIELD_ALBUM_GAIN;
	else if (!strcmp (tag, GST_TAG_ALBUM_PEAK))
		return RB_METADATA_FIELD_ALBUM_PEAK;
	else
		return -1;
}

static const char *
rb_metadata_gst_field_to_gst_tag (RBMetaDataField field)
{
	switch (field)
	{
	case RB_METADATA_FIELD_TITLE:
		return GST_TAG_TITLE;
	case RB_METADATA_FIELD_ARTIST:
		return GST_TAG_ARTIST;
	case RB_METADATA_FIELD_ALBUM:
		return GST_TAG_ALBUM;
	case RB_METADATA_FIELD_GENRE:
		return GST_TAG_GENRE;
	case RB_METADATA_FIELD_COMMENT:
		return GST_TAG_COMMENT;
	case RB_METADATA_FIELD_TRACK_NUMBER:
		return GST_TAG_TRACK_NUMBER;
	case RB_METADATA_FIELD_MAX_TRACK_NUMBER:
		return GST_TAG_TRACK_COUNT;
	case RB_METADATA_FIELD_DISC_NUMBER:
		return GST_TAG_ALBUM_VOLUME_NUMBER;
	case RB_METADATA_FIELD_MAX_DISC_NUMBER:
		return GST_TAG_ALBUM_VOLUME_COUNT;
	case RB_METADATA_FIELD_RELEASE_DATE:
		return GST_TAG_DATE;
	case RB_METADATA_FIELD_DESCRIPTION:
		return GST_TAG_DESCRIPTION;
	case RB_METADATA_FIELD_VERSION:
		return GST_TAG_VERSION;
	case RB_METADATA_FIELD_ISRC:
		return GST_TAG_ISRC;
	case RB_METADATA_FIELD_ORGANIZATION:
		return GST_TAG_ORGANIZATION;
	case RB_METADATA_FIELD_COPYRIGHT:
		return GST_TAG_COPYRIGHT;
	case RB_METADATA_FIELD_CONTACT:
		return GST_TAG_CONTACT;
	case RB_METADATA_FIELD_LICENSE:
		return GST_TAG_LICENSE;
	case RB_METADATA_FIELD_PERFORMER:
		return GST_TAG_PERFORMER;
	case RB_METADATA_FIELD_DURATION:
		return GST_TAG_DURATION;
	case RB_METADATA_FIELD_CODEC:
		return GST_TAG_CODEC;
	case RB_METADATA_FIELD_BITRATE:
		return GST_TAG_BITRATE;
	case RB_METADATA_FIELD_TRACK_GAIN:
		return GST_TAG_TRACK_GAIN;
	case RB_METADATA_FIELD_TRACK_PEAK:
		return GST_TAG_TRACK_PEAK;
	case RB_METADATA_FIELD_ALBUM_GAIN:
		return GST_TAG_ALBUM_GAIN;
	case RB_METADATA_FIELD_ALBUM_PEAK:
		return GST_TAG_ALBUM_PEAK;
	default:
		return NULL;
	}
}

static const char *
rb_metadata_gst_type_to_name (RBMetaData *md, const char *mimetype)
{
	int i;
	for (i = 0; i < md->priv->supported_types->len; i++) {
		struct RBMetadataGstType *type = g_ptr_array_index (md->priv->supported_types, i);
		if (!strcmp (type->mimetype, mimetype))
			return type->human_name;
	}
	return NULL;
}

static RBSetupTaggerElem
rb_metadata_gst_type_to_tag_setup_function (RBMetaData *md, const char *mimetype)
{
	int i;
	for (i = 0; i < md->priv->supported_types->len; i++) {
		struct RBMetadataGstType *type = g_ptr_array_index (md->priv->supported_types, i);
		if (!strcmp (type->mimetype, mimetype))
			return type->setup_tagger_func;
	}
	return NULL;
}

static RBCompleteTaggerElem
rb_metadata_gst_type_to_tag_complete_function (RBMetaData *md, const char *mimetype)
{
	int i;
	for (i = 0; i < md->priv->supported_types->len; i++) {
		struct RBMetadataGstType *type = g_ptr_array_index (md->priv->supported_types, i);
		if (!strcmp (type->mimetype, mimetype))
			return type->complete_tagger_func;
	}
	return NULL;
}

static void
rb_metadata_gst_eos_cb (GstElement *element, RBMetaData *md)
{
	rb_debug ("caught eos");
	if (md->priv->eos) {
		rb_debug ("RENTERED EOS!");
		return;
	}
	gst_element_set_state (md->priv->sink, GST_STATE_NULL);
	md->priv->eos = TRUE;
}

static void
rb_metadata_gst_error_cb (GstElement *element,
			  GstElement *source,
			  GError *error,
			  gchar *debug,
			  RBMetaData *md)
{
	rb_debug ("caught error: %s ", error->message);
	md->priv->error = g_error_new_literal (RB_METADATA_ERROR,
					       RB_METADATA_ERROR_GENERAL,
					       error->message);
}

static void
rb_metadata_gst_load_tag (const GstTagList *list, const gchar *tag, RBMetaData *md)
{
	int count, tem, type;
	RBMetaDataField field;
	GValue *newval;
	const GValue *val;

	rb_debug ("uri: %s tag: %s ", md->priv->uri, tag);

	count = gst_tag_list_get_tag_size (list, tag);
	if (count < 1)
		return;

	tem = rb_metadata_gst_tag_to_field (tag);
	if (tem < 0)
		return;
	field = (RBMetaDataField) tem;

	type = rb_metadata_get_field_type (md, field);
	val = gst_tag_list_get_value_index (list, tag, 0);
	newval = g_new0 (GValue, 1);
	g_value_init (newval, type);
	if (!g_value_transform (val, newval)) {
		
		rb_debug ("Could not transform tag value type %s into %s",
			  g_type_name (G_VALUE_TYPE (val)), 
			  g_type_name (G_VALUE_TYPE (newval)));
		g_value_unset (newval);
		g_free (newval);
		return;
	}

	switch (type) {
	case G_TYPE_STRING: {
		/* Remove leading and trailing whitespace */
		char *str;
		str = g_strdup (g_value_get_string (newval));
		str = g_strstrip (str);
		g_value_take_string (newval, str);
		break;
	}
	default:
		break;
	}

	switch (field) {
	case RB_METADATA_FIELD_BITRATE: {
		/* GStreamer sends us bitrate in bps, but we need it in kbps*/
		gulong bitrate;
		bitrate = g_value_get_ulong (newval);
		g_value_set_ulong (newval, bitrate/1000);
		break;
	}

	case RB_METADATA_FIELD_DURATION: {
		/* GStreamer sends us duration in ns, 
		 * but we need it in seconds
		 */
		guint64 duration;
		duration = g_value_get_uint64 (val);
		g_value_set_ulong (newval, duration/(1000*1000*1000));
		break;
	}
	case RB_METADATA_FIELD_RELEASE_DATE: {
		GDate *date;

		date = g_date_new_julian (g_value_get_ulong (newval));
		g_date_free (date);
		break;
	}
	default:
		break;
	}

	g_hash_table_insert (md->priv->metadata, 
			     GINT_TO_POINTER (field),
			     newval);
}

static void
rb_metadata_gst_found_tag (GObject *pipeline, GstElement *source, GstTagList *tags, RBMetaData *md)
{
	gst_tag_list_foreach (tags, (GstTagForeachFunc) rb_metadata_gst_load_tag, md);
}

static void
rb_metadata_gst_typefind_cb (GstElement *typefind, guint probability, GstCaps *caps, RBMetaData *md)
{
	if (gst_caps_get_size (caps) > 0) {
		md->priv->type = g_strdup (gst_structure_get_name (gst_caps_get_structure (caps, 0)));
		rb_debug ("found type %s", md->priv->type);
	}
}

static void
rb_metadata_gst_fakesink_handoff_cb (GstElement *fakesink, GstBuffer *buf, GstPad *pad, RBMetaData *md)
{
	if (md->priv->handoff) {
		/* we get lots of these with decodebin.  why? */
		/* rb_debug ("caught recursive handoff!"); */
		return;
	} else if (md->priv->eos) {
		rb_debug ("caught handoff after eos!");
		return;
	}
		
	rb_debug ("in fakesink handoff");
	md->priv->handoff = TRUE;
}

static void
rb_metadata_gst_new_decoded_pad_cb (GstElement *decodebin, GstPad *pad, gboolean last, RBMetaData *md)
{
	GstCaps *caps;
	GstStructure *structure;
	const gchar *mimetype;

	caps = gst_pad_get_caps (pad);

	/* we get "ANY" caps for text/plain files etc. */
	if (gst_caps_is_empty (caps) || gst_caps_is_any (caps)) {
		rb_debug ("decoded pad with no caps or any caps.  this file is boring.");
		md->priv->non_audio = TRUE;
	} else {
		/* is this pad audio? */
		structure = gst_caps_get_structure (caps, 0);
		mimetype = gst_structure_get_name (structure);
		if (g_str_has_prefix (mimetype, "audio/x-raw")) {
			/* link this to the fake sink */
			GstPad *sink_pad;
			rb_debug ("linking new decoded pad of type %s to fakesink", mimetype);
			sink_pad = gst_element_get_pad (md->priv->sink, "sink");
			gst_pad_link (pad, sink_pad);

			/* what happens if we get two audio pads from the same stream? 
			 * can this happen?
			 */
		} else {
			/* assume anything we can get a video or text stream out of is
			 * something that should be fed to totem rather than rhythmbox.
			 */
			rb_debug ("got decoded pad of non-audio type %s", mimetype);
			md->priv->non_audio = TRUE;
		}
	}

	gst_caps_free (caps);
}

static void
rb_metadata_gst_unknown_type_cb (GstElement *decodebin, GstPad *pad, GstCaps *caps, RBMetaData *md)
{
	/* try to shortcut it a bit */
	rb_debug ("decodebin emitted unknown type signal");
	md->priv->non_audio = TRUE;
}

static GstElement *make_pipeline_element (GstElement *pipeline, const char *element, GError **error)
{
	GstElement *elem = gst_element_factory_make (element, element);
	if (elem == NULL) {
		g_set_error (error,
			     RB_METADATA_ERROR,
			     RB_METADATA_ERROR_MISSING_PLUGIN,
			     _("Failed to create %s element; check your installation"),
			     element);
		return NULL;
	}

	gst_bin_add (GST_BIN (pipeline), elem);
	return elem;
}

void
rb_metadata_load (RBMetaData *md,
		  const char *uri,
		  GError **error)
{
	GstElement *pipeline = NULL;
	GstElement *gnomevfssrc = NULL;
	GstElement *decodebin = NULL;
	GstElement *typefind = NULL;

	g_free (md->priv->uri);
	md->priv->uri = NULL;
	g_free (md->priv->type);
	md->priv->type = NULL;
	md->priv->error = NULL;
	md->priv->eos = FALSE;
	md->priv->handoff = FALSE;
	md->priv->non_audio = FALSE;

	if (md->priv->pipeline) {
		gst_object_unref (GST_OBJECT (md->priv->pipeline));
		md->priv->pipeline = NULL;
	}

	if (uri == NULL)
		return;

	rb_debug ("loading metadata for uri: %s", uri);
	md->priv->uri = g_strdup (uri);

	if (md->priv->metadata)
		g_hash_table_destroy (md->priv->metadata);
	md->priv->metadata = g_hash_table_new_full (g_direct_hash, g_direct_equal,
						    NULL, (GDestroyNotify) free_gvalue);

	/* The main tagfinding pipeline looks like this:
	 * gnomevfssrc ! decodebin ! fakesink
	 *
	 * but we can only link the fakesink in when the decodebin
	 * creates an audio source pad.  we do this in the 'new-decoded-pad'
	 * signal handler.
	 */
	pipeline = gst_pipeline_new ("pipeline");

	gnomevfssrc = make_pipeline_element (pipeline, "gnomevfssrc", error);
	decodebin = make_pipeline_element (pipeline, "decodebin", error);
	md->priv->sink = make_pipeline_element (pipeline, "fakesink", error);
	if (!(gnomevfssrc && decodebin && md->priv->sink)) {
		rb_debug ("missing an element, sadly");
		goto out;
	}

	g_object_set (G_OBJECT (gnomevfssrc), "location", uri, NULL);
	g_object_set (G_OBJECT (md->priv->sink), "signal-handoffs", TRUE, NULL);

	g_signal_connect_object (pipeline, "error", G_CALLBACK (rb_metadata_gst_error_cb), md, 0);
	g_signal_connect_object (pipeline, "found-tag", G_CALLBACK (rb_metadata_gst_found_tag), md, 0);
	
	g_signal_connect_object (md->priv->sink, "handoff", G_CALLBACK (rb_metadata_gst_fakesink_handoff_cb), md, 0);
	g_signal_connect_object (md->priv->sink, "eos", G_CALLBACK (rb_metadata_gst_eos_cb), md, 0);

	g_signal_connect_object (decodebin, "new-decoded-pad", G_CALLBACK (rb_metadata_gst_new_decoded_pad_cb), md, 0);
	g_signal_connect_object (decodebin, "unknown-type", G_CALLBACK (rb_metadata_gst_unknown_type_cb), md, 0);

	/* locate the decodebin's typefind, so we can get the have_type signal too.
	 * this is kind of nasty, since it relies on an essentially arbitrary string 
	 * in the decodebin code not changing.  the alternative is to have our own
	 * typefind instance before the decodebin.  it might not like that.
	 */
	typefind = gst_bin_get_by_name (GST_BIN (decodebin), "typefind");
	g_assert (typefind != NULL);
	g_signal_connect_object (typefind, "have_type", G_CALLBACK (rb_metadata_gst_typefind_cb), md, 0);

	gst_element_link (gnomevfssrc, decodebin);

	md->priv->pipeline = pipeline;
	gst_element_set_state (pipeline, GST_STATE_PLAYING);
	while (gst_bin_iterate (GST_BIN (pipeline))
	       && md->priv->error == NULL
	       && !md->priv->non_audio
	       && !md->priv->handoff
	       && !md->priv->eos)
		;

	if (md->priv->handoff) {
		/* We caught the first buffer, which means the decoder should have read all
		 * of the metadata, and should know the length now.
		 */
		if (g_hash_table_lookup (md->priv->metadata, GINT_TO_POINTER (RB_METADATA_FIELD_DURATION)) == NULL) {
			GstFormat format = GST_FORMAT_TIME;
			gint64 length;
			
			if (gst_element_query (md->priv->sink, GST_QUERY_TOTAL,
					       &format, &length)) {
				GValue *newval = g_new0 (GValue, 1);
				
				rb_debug ("duration query succeeded");
				
				g_value_init (newval, G_TYPE_ULONG);
				/* FIXME - use guint64 for duration? */
				g_value_set_ulong (newval, (long) (length / (1 * 1000 * 1000 * 1000)));
				g_hash_table_insert (md->priv->metadata, GINT_TO_POINTER (RB_METADATA_FIELD_DURATION),
						     newval);
			} else {
				rb_debug ("duration query failed!");
			}
		}
	} else if (md->priv->eos)
		rb_debug ("caught eos without handoff!");
	
	gst_element_set_state (pipeline, GST_STATE_NULL);

	/* report errors for various failure cases.
	 * these don't include the URI as the load failure dialog
	 * already displays it.
	 */
	if (md->priv->error) {
		g_propagate_error (error, md->priv->error);
	} else if (!md->priv->type) {
		g_set_error (error,
			     RB_METADATA_ERROR,
			     RB_METADATA_ERROR_UNRECOGNIZED,
			     _("The MIME type of the file could not be identified"));
	} else if (md->priv->non_audio) {
		g_set_error (error,
			     RB_METADATA_ERROR,
			     RB_METADATA_ERROR_NOT_AUDIO,
			     _("The file is not an audio stream"));
	} else if (!md->priv->handoff) {
		const char *report_type = rb_metadata_gst_type_to_name (md, md->priv->type);
		if (report_type == NULL)
			report_type = md->priv->type;
		g_set_error (error,
			     RB_METADATA_ERROR,
			     RB_METADATA_ERROR_UNSUPPORTED,
			     _("There is no plugin installed to handle a %s file."),
			     md->priv->type);
	} else {
		/* yay, it worked */
		rb_debug ("successfully read metadata for %s", uri);
	
		/* it doesn't matter if we don't recognise the format,
		 * as long as gstreamer can play it, we can put it in
		 * the library.
		 */
		if (!rb_metadata_gst_type_to_name (md, md->priv->type)) {
			rb_debug ("we don't know what type %s (from file %s) is, but we'll play it anyway",
				  md->priv->type, uri);
		}
	}

	if (*error != NULL) {
		g_free (md->priv->type);
		md->priv->type = NULL;
	}
 out:
	if (pipeline != NULL)
		gst_object_unref (GST_OBJECT (pipeline));
	md->priv->pipeline = NULL;
}

gboolean
rb_metadata_can_save (RBMetaData *md, const char *mimetype)
{
#ifdef ENABLE_TAG_WRITING
	return rb_metadata_gst_type_to_tag_setup_function (md, mimetype) != NULL;
#else 
	return FALSE;
#endif
}

static void
rb_metadata_gst_add_tag_data (gpointer key, const GValue *val, GstTagSetter *tagsetter)
{
	RBMetaDataField field = GPOINTER_TO_INT (key);
	const char *tag = rb_metadata_gst_field_to_gst_tag (field);

	if (tag) {
		GValue newval = {0,};
		g_value_init (&newval, gst_tag_get_type (tag));
		if (g_value_transform (val, &newval)) {
			gst_tag_setter_add_values (GST_TAG_SETTER (tagsetter),
						   GST_TAG_MERGE_REPLACE,
						   tag, &newval, NULL);
		}
		g_value_unset (&newval);
	}
}



void
rb_metadata_save (RBMetaData *md, GError **error)
{
	GstElement *pipeline = NULL;
	GstElement *gnomevfssrc = NULL;
	GstElement *tagger = NULL;
	GstElement *tagger_link = NULL;
	const char *plugin_name = NULL;
	char *tmpname = NULL;
	GnomeVFSHandle *handle = NULL;
	GnomeVFSResult result;
	RBSetupTaggerElem setup_tagger_func;
	RBCompleteTaggerElem complete_tagger_func;
	RBTaggerSetupData setup_data;

	g_return_if_fail (md->priv->uri != NULL);
	g_return_if_fail (md->priv->type != NULL);

	rb_debug ("saving metadata for uri: %s", md->priv->uri);

	if ((result = rb_uri_mkstemp (md->priv->uri, &tmpname, &handle)) != GNOME_VFS_OK)
		goto vfs_error;

	pipeline = gst_pipeline_new ("pipeline");

	g_signal_connect_object (pipeline, "error", G_CALLBACK (rb_metadata_gst_error_cb), md, 0);

	/* Source */
	plugin_name = "gnomevfssrc";
	if (!(gnomevfssrc = gst_element_factory_make (plugin_name, plugin_name)))
		goto missing_plugin;
	gst_bin_add (GST_BIN (pipeline), gnomevfssrc);		  
	g_object_set (G_OBJECT (gnomevfssrc), "location", md->priv->uri, NULL);

	setup_tagger_func = rb_metadata_gst_type_to_tag_setup_function (md, md->priv->type);
	complete_tagger_func = rb_metadata_gst_type_to_tag_complete_function (md, md->priv->type);
	
	/* Set up tagger pipeline construction */
	if (!(setup_tagger_func && complete_tagger_func)) {
		g_set_error (error,
			     RB_METADATA_ERROR,
			     RB_METADATA_ERROR_UNSUPPORTED,
			     "Unsupported file type: %s", md->priv->type);
		goto out_error;
	}

	setup_data.pad = NULL;
	setup_data.complete = FALSE;
	if (!setup_tagger_func (GST_BIN (pipeline), gnomevfssrc, &setup_data)) {
		g_set_error (error,
			     RB_METADATA_ERROR,
			     RB_METADATA_ERROR_UNSUPPORTED,
			     "Unable to construct pipeline to tag %s files", md->priv->type);
		goto out_error;
	}

	/* Iterate until we can fully construct the pipeline */
	if (setup_data.complete == FALSE) {
		gst_element_set_state (pipeline, GST_STATE_PLAYING);
		while (setup_data.complete == FALSE && gst_bin_iterate (GST_BIN (pipeline)) && md->priv->error == NULL)
			;
		gst_element_set_state (pipeline, GST_STATE_PAUSED);
	}

	/* Finish constructing the pipeline */
	if (!complete_tagger_func (GST_BIN (pipeline), &setup_data, &tagger, &tagger_link)) {
		g_set_error (error,
			     RB_METADATA_ERROR,
			     RB_METADATA_ERROR_UNSUPPORTED,
			     "Unable to construct pipeline to tag %s files", md->priv->type);
		goto out_error;
	}

	gst_tag_setter_set_merge_mode (GST_TAG_SETTER (tagger), GST_TAG_MERGE_REPLACE);
	g_hash_table_foreach (md->priv->metadata, 
			      (GHFunc) rb_metadata_gst_add_tag_data,
			      GST_TAG_SETTER (tagger));

	/* Sink */
	plugin_name = "gnomevfssink";
	if (!(md->priv->sink = gst_element_factory_make (plugin_name, plugin_name)))
		goto missing_plugin;

	g_object_set (G_OBJECT (md->priv->sink), "handle", handle, NULL);
	g_signal_connect_object (md->priv->sink, "eos", G_CALLBACK (rb_metadata_gst_eos_cb), md, 0);
	gst_bin_add (GST_BIN (pipeline), md->priv->sink); 

	gst_element_link_many (tagger_link, md->priv->sink, NULL);

	md->priv->pipeline = pipeline;
	gst_element_set_state (pipeline, GST_STATE_PLAYING);
	while (gst_bin_iterate (GST_BIN (pipeline))
	       && md->priv->error == NULL
	       && !md->priv->eos)
		;
	gst_element_set_state (pipeline, GST_STATE_NULL);

	if (md->priv->error) {
		g_propagate_error (error, md->priv->error);
		goto out_error;
	}

	if (handle != NULL) {
		GnomeVFSFileInfo info;
		if ((result = gnome_vfs_close (handle)) != GNOME_VFS_OK)
			goto vfs_error;
		handle = NULL;
		if ((result = gnome_vfs_get_file_info (tmpname, &info, 0)) != GNOME_VFS_OK)
			goto vfs_error;
		if (info.size == 0) {
			/* TODO report error properly... */
			goto out_error;
		}
		/* XXX test
		if ((result = gnome_vfs_move (tmpname, md->priv->uri, TRUE)) != GNOME_VFS_OK)
				goto vfs_error;
		*/
	}

	goto out;
 vfs_error:
	g_set_error (error,
		     RB_METADATA_ERROR,
		     RB_METADATA_ERROR_GNOMEVFS,
		     "%s",
		     gnome_vfs_result_to_string (result)); 
	goto out_error;
 missing_plugin:
	g_set_error (error,
		     RB_METADATA_ERROR,
		     RB_METADATA_ERROR_MISSING_PLUGIN,
		     _("Failed to create %s element; check your installation"),
		     plugin_name); 
 out_error:
	if (handle != NULL)
		gnome_vfs_close (handle);
	if (tmpname != NULL)
		gnome_vfs_unlink (tmpname);
 out:
	if (pipeline != NULL)
		gst_object_unref (GST_OBJECT (pipeline));
	md->priv->pipeline = NULL;
}

gboolean
rb_metadata_get (RBMetaData *md, RBMetaDataField field,
		 GValue *ret)
{
	GValue *val;
	if ((val = g_hash_table_lookup (md->priv->metadata,
					GINT_TO_POINTER (field)))) {
		g_value_init (ret, G_VALUE_TYPE (val));
		g_value_copy (val, ret);
		return TRUE;
	}
	return FALSE;
}

const char *
rb_metadata_get_mime (RBMetaData *md)
{
	return md->priv->type;
}

gboolean
rb_metadata_set (RBMetaData *md, RBMetaDataField field,
		 const GValue *val)
{
	GValue *newval;
	GType type;
	
	if (rb_metadata_gst_field_to_gst_tag (field) == NULL) {
		return FALSE;
	}

	type = rb_metadata_get_field_type (md, field);
	g_return_val_if_fail (type == G_VALUE_TYPE (val), FALSE);

	newval = g_new0 (GValue, 1);
	g_value_init (newval, type);
	g_value_copy (val, newval);

	g_hash_table_insert (md->priv->metadata, GINT_TO_POINTER (field),
			     newval);
	return TRUE;
}

