/*
 *  arch-tag: Implementation of Rhythmbox DBus remoting
 *
 *  Copyright (C) 2004 Colin Walters <walters@gnome.org>
 *  		  2005 Jonathan Matthew <jonathan@kaolin.hn.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <config.h>
#include "rb-remote-dbus.h"
#include "rb-remote-client-proxy.h"
#include <libgnome/libgnome.h>
#include <libgnome/gnome-i18n.h>
#include <string.h>
#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>
#include <dbus/dbus-glib-bindings.h>
#include <gdk/gdk.h>

#include "org_gnome_Rhythmbox-glue.h"
#include "org_gnome_Rhythmbox-binding.h"

#include "rb-debug.h"

#define RB_REMOTE_DBUS_SERVICE_PATH "org.gnome.Rhythmbox"
#define RB_REMOTE_DBUS_NAME "org.gnome.Rhythmbox"
#define RB_REMOTE_DBUS_OBJECT_PATH "/org/gnome/Rhythmbox/DBusRemote"

static void rb_remote_dbus_remote_client_proxy_init (RBRemoteClientProxyIface *iface);

G_DEFINE_TYPE_WITH_CODE(RBRemoteDBus, rb_remote_dbus, G_TYPE_OBJECT,
			G_IMPLEMENT_INTERFACE(RB_TYPE_REMOTE_CLIENT_PROXY,
					      rb_remote_dbus_remote_client_proxy_init))

static void rb_remote_dbus_create_client_proxy (RBRemoteDBus *dbus);

static void rb_remote_dbus_dispose (GObject *object);
static void rb_remote_dbus_finalize (GObject *object);

static void rb_remote_dbus_play_order_notify_cb (GObject *object,
						 GParamSpec *param,
						 RBRemoteDBus *dbus);

/* Client methods */
static void rb_remote_dbus_client_handle_uri_impl (RBRemoteClientProxy *proxy, const char *uri);
static RBRemoteSong *rb_remote_dbus_client_get_playing_song_impl (RBRemoteClientProxy *proxy);
static void rb_remote_dbus_client_grab_focus_impl (RBRemoteClientProxy *proxy);
static void rb_remote_dbus_client_toggle_shuffle_impl (RBRemoteClientProxy *proxy);
static void rb_remote_dbus_client_toggle_repeat_impl (RBRemoteClientProxy *proxy);
static void rb_remote_dbus_client_toggle_playing_impl (RBRemoteClientProxy *proxy);
static long rb_remote_dbus_client_get_playing_time_impl (RBRemoteClientProxy *proxy);
static void rb_remote_dbus_client_set_playing_time_impl (RBRemoteClientProxy *proxy, long time);
static void rb_remote_dbus_client_quit_impl (RBRemoteClientProxy *proxy);

static GObjectClass *parent_class;

enum
{
	PROP_REPEAT,
	PROP_SHUFFLE,
	PROP_NONE,
};

enum
{
	SHUFFLE_CHANGED,
	REPEAT_CHANGED,
	PLAYING_CHANGED,
	PLAYING_SONG_CHANGED,
	LAST_SIGNAL,
};

struct RBRemoteDBusPrivate
{
	gboolean disposed;

	DBusGConnection *connection;
	DBusGProxy *dbus_proxy;

	/* used to track changes */
	gboolean shuffle;
	gboolean repeat;
	
	RBRemoteProxy *proxy;

	GHashTable *handlers;
};

static guint rb_remote_dbus_signals[LAST_SIGNAL] = { 0 };

static void
rb_remote_dbus_class_init (RBRemoteDBusClass *klass)
{
        GObjectClass *object_class = (GObjectClass *) klass;

        parent_class = g_type_class_peek_parent (klass);

        object_class->dispose = rb_remote_dbus_dispose;
        object_class->finalize = rb_remote_dbus_finalize;

	rb_remote_dbus_signals[SHUFFLE_CHANGED] =
		g_signal_new ("shuffle-changed",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (RBRemoteDBusClass, shuffle_changed),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__BOOLEAN,
			      G_TYPE_NONE,
			      1,
			      G_TYPE_BOOLEAN);
	
	rb_remote_dbus_signals[REPEAT_CHANGED] =
		g_signal_new ("repeat-changed",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (RBRemoteDBusClass, repeat_changed),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__BOOLEAN,
			      G_TYPE_NONE,
			      1,
			      G_TYPE_BOOLEAN);
	
	rb_remote_dbus_signals[PLAYING_CHANGED] =
		g_signal_new ("playing-changed",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (RBRemoteDBusClass, playing_changed),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__BOOLEAN,
			      G_TYPE_NONE,
			      1,
			      G_TYPE_BOOLEAN);
	
	dbus_g_object_class_install_info (G_OBJECT_CLASS (object_class),
					  &dbus_glib_rb_remote_dbus_object_info);
}

static void
rb_remote_dbus_remote_client_proxy_init (RBRemoteClientProxyIface *iface)
{
	iface->handle_uri = rb_remote_dbus_client_handle_uri_impl;
	iface->get_playing_song = rb_remote_dbus_client_get_playing_song_impl;
	iface->grab_focus = rb_remote_dbus_client_grab_focus_impl;
	iface->toggle_shuffle = rb_remote_dbus_client_toggle_shuffle_impl;
	iface->toggle_repeat = rb_remote_dbus_client_toggle_repeat_impl;
	iface->toggle_playing = rb_remote_dbus_client_toggle_playing_impl;
	iface->get_playing_time = rb_remote_dbus_client_get_playing_time_impl;
	iface->set_playing_time = rb_remote_dbus_client_set_playing_time_impl;
	iface->quit = rb_remote_dbus_client_quit_impl;
}

static void
emit_boolean_signal (DBusGConnection *connection, const char *path, const char *interface, const char *signal, gboolean value)
{
	/* eww, like, manual labour? */
	DBusMessageIter iter;
	DBusMessage *message = dbus_message_new_signal (RB_REMOTE_DBUS_OBJECT_PATH,
							interface, signal);
	dbus_message_iter_init_append (message, &iter);
	dbus_message_iter_append_basic (&iter, DBUS_TYPE_BOOLEAN, &value);
	
	if (!dbus_connection_send (dbus_g_connection_get_connection (connection), message, NULL))
		rb_debug ("dbus_connection_send failed?");

	dbus_g_connection_flush (connection);

	dbus_message_unref (message);
}

static void
emit_shuffle_changed_cb (RBRemoteDBus *dbus, gboolean shuffle, gpointer blah)
{
	emit_boolean_signal (dbus->priv->connection, 
			     RB_REMOTE_DBUS_OBJECT_PATH, 
			     "org.gnome.Rhythmbox", 
			     "ShuffleChanged", 
			     shuffle);
}

static void
emit_repeat_changed_cb (RBRemoteDBus *dbus, gboolean repeat, gpointer blah)
{
	emit_boolean_signal (dbus->priv->connection, 
			     RB_REMOTE_DBUS_OBJECT_PATH, 
			     "org.gnome.Rhythmbox", 
			     "RepeatChanged", 
			     repeat);
}


static void
rb_remote_dbus_init (RBRemoteDBus *dbus) 
{
	GError *error = NULL;
	dbus->priv = g_new0 (RBRemoteDBusPrivate, 1);

	dbus->priv->connection = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
	if (dbus->priv->connection == NULL) {
		g_critical ("couldn't connect to session bus: %s",
			    error->message);
		g_error_free (error);
		return;
	}

	if (error != NULL)
		g_error_free (error);

	/* hopefully temporary: manually convert our signals into dbus messages */
	g_signal_connect (G_OBJECT (dbus), "shuffle-changed", G_CALLBACK (emit_shuffle_changed_cb), NULL);
	g_signal_connect (G_OBJECT (dbus), "repeat-changed", G_CALLBACK (emit_repeat_changed_cb), NULL);
}

static void
rb_remote_dbus_dispose (GObject *object)
{
	RBRemoteDBus *dbus = RB_REMOTE_DBUS (object);

	if (dbus->priv->disposed)
		return;
	dbus->priv->disposed = TRUE;
	
	if (dbus->priv->connection) {
		dbus_g_connection_unref (dbus->priv->connection);
		dbus->priv->connection = NULL;
	}
}

static void
rb_remote_dbus_finalize (GObject *object)
{
        RBRemoteDBus *dbus = RB_REMOTE_DBUS (object);

	g_free (dbus->priv);

	(* G_OBJECT_CLASS (parent_class)->finalize) (object);
}

RBRemoteDBus *
rb_remote_dbus_new (void)
{
	return g_object_new (RB_TYPE_REMOTE_DBUS, NULL);
}

GQuark
rb_remote_dbus_error_quark (void)
{
	static GQuark quark = 0;
	if (!quark)
		quark = g_quark_from_static_string ("rb_remote_dbus_error");

	return quark;
}

static void
rb_remote_dbus_play_order_notify_cb (GObject *object,
				     GParamSpec *param,
				     RBRemoteDBus *dbus)
{
	gboolean new_shuffle;
	gboolean new_repeat;

	GValue value = {0};
	g_value_init (&value, G_PARAM_SPEC_VALUE_TYPE (param));
	new_shuffle = rb_remote_proxy_get_shuffle (dbus->priv->proxy);
	new_repeat = rb_remote_proxy_get_repeat (dbus->priv->proxy);

	if (new_shuffle != dbus->priv->shuffle) {
		g_signal_emit (G_OBJECT (dbus), rb_remote_dbus_signals[SHUFFLE_CHANGED], 0, new_shuffle);
		dbus->priv->shuffle = new_shuffle;
	}
	if (new_repeat != dbus->priv->repeat) {
		g_signal_emit (G_OBJECT (dbus), rb_remote_dbus_signals[REPEAT_CHANGED], 0, new_repeat);
		dbus->priv->repeat = new_repeat;
	}
}

gboolean
rb_remote_dbus_activate (RBRemoteDBus *dbus)
{
	GError *error = NULL;
	DBusGProxy *bus_proxy;
	gboolean owner;
	gboolean ret;

	if (!dbus->priv->connection)
		return FALSE;

	bus_proxy = dbus_g_proxy_new_for_name (dbus->priv->connection,
					       DBUS_SERVICE_DBUS,
					       DBUS_PATH_DBUS,
					       DBUS_INTERFACE_DBUS);
	ret = org_freedesktop_DBus_name_has_owner (bus_proxy,
						   g_strdup (RB_REMOTE_DBUS_NAME),
						   &owner,
						   &error);
	g_object_unref (bus_proxy);
	if (ret == FALSE) {
		rb_debug ("Couldn't check name " RB_REMOTE_DBUS_NAME ": %s", error->message);
		g_error_free (error);
		return FALSE;
	}

	return owner;
}

static void
rb_remote_dbus_create_client_proxy (RBRemoteDBus *dbus)
{
	dbus->priv->dbus_proxy = dbus_g_proxy_new_for_name (dbus->priv->connection,
							    RB_REMOTE_DBUS_NAME,
							    RB_REMOTE_DBUS_OBJECT_PATH,
							    RB_REMOTE_DBUS_SERVICE_PATH);

	/* TODO connect signals */
}

static void
shell_weak_ref_cb (gpointer data, GObject *objptr)
{
	g_object_unref (G_OBJECT (data));
}

gboolean
rb_remote_dbus_acquire (RBRemoteDBus *dbus,
			RBRemoteProxy *proxy,
			GError **error)
{
	DBusGProxy *bus_proxy;
	gboolean ret;
	guint32 request_name_ret;

	if (!dbus->priv->connection)
		return FALSE;
	
	dbus->priv->proxy = proxy;
	g_object_weak_ref (G_OBJECT (proxy), shell_weak_ref_cb, dbus);


	dbus_g_connection_register_g_object (dbus->priv->connection,
					     RB_REMOTE_DBUS_OBJECT_PATH,
					     G_OBJECT (dbus));
	
	bus_proxy = dbus_g_proxy_new_for_name (dbus->priv->connection,
					       DBUS_SERVICE_DBUS,
					       DBUS_PATH_DBUS,
					       DBUS_INTERFACE_DBUS);
	ret = org_freedesktop_DBus_request_name (bus_proxy,
						 RB_REMOTE_DBUS_NAME, 0,
						 &request_name_ret, error);
	g_object_unref (bus_proxy);
	if (ret == FALSE)
		return FALSE;
	
	if (request_name_ret != DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER) {
		g_set_error (error,
			     RB_REMOTE_DBUS_ERROR,
			     RB_REMOTE_DBUS_ERROR_REQUEST_NAME_FAILURE,
			     "Got result code %u from requesting name",
			     request_name_ret);
		return FALSE;
	}

	rb_debug ("DBus remote interface has acquired name " RB_REMOTE_DBUS_NAME);

	dbus->priv->shuffle = rb_remote_proxy_get_shuffle (dbus->priv->proxy);
	dbus->priv->repeat = rb_remote_proxy_get_repeat (dbus->priv->proxy);

	/* connect to signals from player object */
	rb_remote_proxy_player_notify_handler (dbus->priv->proxy,
					       "play-order",
					       G_CALLBACK (rb_remote_dbus_play_order_notify_cb),
					       dbus);
	return TRUE;
}

/* XXX can we do this with the glib bindings?
static void
rb_remote_dbus_unregister_handler (DBusConnection *connection, void *data)
{
	rb_debug ("unregistered!");
}
*/

gboolean
rb_remote_dbus_quit (RBRemoteDBus *dbus, GError **error)
{
	rb_remote_proxy_quit (dbus->priv->proxy);
	return TRUE;
}

gboolean
rb_remote_dbus_grab_focus (RBRemoteDBus *dbus, GError **error)
{
	rb_remote_proxy_grab_focus (dbus->priv->proxy);
	return TRUE;
}

gboolean
rb_remote_dbus_handle_uri (RBRemoteDBus *dbus, const char *uri, GError **error)
{
	/* XXX do something */
	return TRUE;
}

gboolean
rb_remote_dbus_get_playing (RBRemoteDBus *dbus, gboolean *playing, GError **error)
{
	*playing = rb_remote_proxy_playing (dbus->priv->proxy);
	return TRUE;
}

gboolean
rb_remote_dbus_play_pause (RBRemoteDBus *dbus, GError **error)
{
        if (rb_remote_proxy_playing (dbus->priv->proxy))
                rb_remote_proxy_pause (dbus->priv->proxy);
        else
                rb_remote_proxy_play (dbus->priv->proxy);
	return TRUE;
}

gboolean
rb_remote_dbus_play (RBRemoteDBus *dbus, GError **error)
{
	rb_remote_proxy_play (dbus->priv->proxy);
	return TRUE;
}

gboolean
rb_remote_dbus_pause (RBRemoteDBus *dbus, GError **error)
{
	rb_remote_proxy_pause (dbus->priv->proxy);
	return TRUE;
}

gboolean
rb_remote_dbus_jump_next (RBRemoteDBus *dbus, GError **error)
{
	rb_remote_proxy_jump_next (dbus->priv->proxy);
	return TRUE;
}

gboolean
rb_remote_dbus_jump_previous (RBRemoteDBus *dbus, GError **error)
{
	rb_remote_proxy_jump_previous (dbus->priv->proxy);
	return TRUE;
}

gboolean
rb_remote_dbus_get_playing_time (RBRemoteDBus *dbus, gulong *time, GError **error)
{
	*time = rb_remote_proxy_get_playing_time (dbus->priv->proxy);
	return TRUE;
}

gboolean
rb_remote_dbus_set_playing_time (RBRemoteDBus *dbus, gulong time, GError **error)
{
	rb_remote_proxy_set_playing_time (dbus->priv->proxy, time);
	return TRUE;
}

gboolean
rb_remote_dbus_get_playing_song_uri (RBRemoteDBus *dbus, char **song_uri, guint32 *pants, GError **error)
{
	gchar *uri = rb_remote_proxy_get_playing_uri (dbus->priv->proxy);
	if (uri == NULL)
		uri = "";		// can't return NULLs through DBus?
	*song_uri = g_strdup(uri);
	*pants = 37;
	return TRUE;
}

gboolean
rb_remote_dbus_get_song_info (RBRemoteDBus *dbus, const char *song_uri, char **title, 
			      char **artist, char **genre, char **album, 
			      long *track_number, long *duration, long *bitrate, 
			      long *filesize, long *rating, long *play_count, 
			      long *last_played, long *disc_number, GError **error)
{
        RBRemoteSong song;
	memset (&song, 0, sizeof (RBRemoteSong));

	*title = g_strdup ("");
	*artist = g_strdup ("");
	*genre = g_strdup ("");
	*album = g_strdup ("");
	*track_number = 0;
	*duration = 0;
	*bitrate = 0;
	*filesize = 0;
	*rating = 0;
	*play_count = 0;
	*last_played = 0;
	*disc_number = 0;

        if (rb_remote_proxy_get_song_info (dbus->priv->proxy, g_strdup (song_uri), &song)) {
		if (song.title != NULL) {
			g_free (*title);
			*title = g_strdup (song.title);
		}
		if (song.artist != NULL) {
			g_free (*artist);
			*artist = g_strdup (song.artist);
		}
		if (song.genre != NULL) {
			g_free (*genre);
			*genre = g_strdup (song.genre);
		}
		if (song.album != NULL) {
			g_free (*album);
			*album = g_strdup (song.album);
		}
		*track_number = song.track_number;
		*duration = song.duration;
		*bitrate = song.bitrate;
		*filesize = song.filesize;
		*rating = song.rating;
		*play_count = song.play_count;
		*last_played = song.last_played;
		*disc_number = song.disc_number;
	}
        
        rb_remote_song_free (&song);
        return TRUE;
}

gboolean
rb_remote_dbus_get_shuffle (RBRemoteDBus *dbus, gboolean *shuffle, GError **error)
{
	*shuffle = rb_remote_proxy_get_shuffle (dbus->priv->proxy);
	return TRUE;
}

gboolean 
rb_remote_dbus_set_shuffle (RBRemoteDBus *dbus, gboolean shuffle, GError **error)
{
	rb_remote_proxy_set_shuffle (dbus->priv->proxy, shuffle);
	return TRUE;
}

gboolean
rb_remote_dbus_toggle_shuffle (RBRemoteDBus *dbus, GError **error)
{
	gboolean shuffle = rb_remote_proxy_get_shuffle (dbus->priv->proxy);
	rb_remote_proxy_set_shuffle (dbus->priv->proxy, !shuffle);
	return TRUE;
}

gboolean
rb_remote_dbus_get_repeat (RBRemoteDBus *dbus, gboolean *repeat, GError **error)
{
	*repeat = rb_remote_proxy_get_repeat (dbus->priv->proxy);
	return TRUE;
}

gboolean 
rb_remote_dbus_set_repeat (RBRemoteDBus *dbus, gboolean repeat, GError **error)
{
	rb_remote_proxy_set_repeat (dbus->priv->proxy, repeat);
	return TRUE;
}

gboolean
rb_remote_dbus_toggle_repeat (RBRemoteDBus *dbus, GError **error)
{
	gboolean repeat = rb_remote_proxy_get_repeat (dbus->priv->proxy);
	rb_remote_proxy_set_repeat (dbus->priv->proxy, !repeat);
	return TRUE;
}



/* Client methods */

static void
rb_remote_dbus_client_handle_uri_impl (RBRemoteClientProxy *proxy, const char *uri)
{
	RBRemoteDBus *dbus = RB_REMOTE_DBUS (proxy);
	GError *error = NULL;
	
	if (dbus->priv->dbus_proxy == NULL)
		rb_remote_dbus_create_client_proxy (dbus);
	
	if (!org_gnome_Rhythmbox_handle_ur_i /* heh */ (dbus->priv->dbus_proxy, uri, &error)) {
		g_error ("Error invoking org_gnome_Rhythmbox_handle_ur_i: %s", error->message);
		g_error_free (error);
	}
}

static RBRemoteSong *
rb_remote_dbus_client_get_playing_song_impl (RBRemoteClientProxy *proxy)
{
	RBRemoteDBus *dbus = RB_REMOTE_DBUS (proxy);
	char *title;
	char *artist;
	char *genre;
	char *album;
	char *uri;
	guint32 track_number;
	guint32 duration;
	guint32 bitrate;
	guint32 filesize;
	guint32 rating;
	guint32 play_count;
	guint32 last_played;
	guint32 disc_number;
	GError *error = NULL;
	RBRemoteSong *song = NULL;
	
	if (dbus->priv->dbus_proxy == NULL)
		rb_remote_dbus_create_client_proxy (dbus);
	
	if (!org_gnome_Rhythmbox_get_playing_song_ur_i (dbus->priv->dbus_proxy, &uri, &error)) {
		g_error ("Error invoking org_gnome_Rhythmbox_get_playing_song_uri: %s", error->message);
		g_error_free (error);
		return NULL;
	}

	if (uri == NULL || strlen(uri) == 0) {
		return NULL;
	}

	if (!org_gnome_Rhythmbox_get_song_info (dbus->priv->dbus_proxy, uri, &title,
						&artist, &genre, &album,
						&track_number, &duration,
						&bitrate, &filesize, &rating, 
						&play_count, &last_played, &disc_number,
						&error)) {
		g_error ("Error invoking org_gnome_Rhythmbox_get_song_info: %s", error->message);
		g_error_free (error);
		return NULL;
	}

	song = g_new0 (RBRemoteSong, 1);
	song->title = g_strdup(title);
	song->artist = g_strdup(artist);
	song->genre = g_strdup(genre);
	song->album = g_strdup(album);
	song->uri = g_strdup(uri);
	song->track_number = track_number;
	song->duration = duration;
	song->bitrate = bitrate;
	song->filesize = filesize;
	song->rating = rating;
	song->play_count = play_count;
	song->last_played = last_played;
	song->disc_number = disc_number;

	/* free things? */
	return song;
}

static void
rb_remote_dbus_client_grab_focus_impl (RBRemoteClientProxy *proxy)
{
	RBRemoteDBus *dbus = RB_REMOTE_DBUS (proxy);
	GError *error = NULL;
	
	if (dbus->priv->dbus_proxy == NULL)
		rb_remote_dbus_create_client_proxy (dbus);
	
	if (!org_gnome_Rhythmbox_grab_focus (dbus->priv->dbus_proxy, &error)) {
		g_error ("Error invoking org_gnome_Rhythmbox_grab_focus: %s", error->message);
		g_error_free (error);
	}
}

static void
rb_remote_dbus_client_toggle_shuffle_impl (RBRemoteClientProxy *proxy)
{
	RBRemoteDBus *dbus = RB_REMOTE_DBUS (proxy);
	GError *error = NULL;
	
	if (dbus->priv->dbus_proxy == NULL)
		rb_remote_dbus_create_client_proxy (dbus);
	
	if (!org_gnome_Rhythmbox_toggle_shuffle (dbus->priv->dbus_proxy, &error)) {
		g_error ("Error invoking org_gnome_Rhythmbox_toggle_shuffle: %s", error->message);
		g_error_free (error);
	}
}

static void
rb_remote_dbus_client_toggle_repeat_impl (RBRemoteClientProxy *proxy)
{
	RBRemoteDBus *dbus = RB_REMOTE_DBUS (proxy);
	GError *error = NULL;
	
	if (dbus->priv->dbus_proxy == NULL)
		rb_remote_dbus_create_client_proxy (dbus);
	
	if (!org_gnome_Rhythmbox_toggle_repeat (dbus->priv->dbus_proxy, &error)) {
		g_error ("Error invoking org_gnome_Rhythmbox_toggle_repeat: %s", error->message);
		g_error_free (error);
	}
}

static void
rb_remote_dbus_client_toggle_playing_impl (RBRemoteClientProxy *proxy)
{
	RBRemoteDBus *dbus = RB_REMOTE_DBUS (proxy);
	GError *error = NULL;
	
	if (dbus->priv->dbus_proxy == NULL)
		rb_remote_dbus_create_client_proxy (dbus);
	
	if (!org_gnome_Rhythmbox_play_pause (dbus->priv->dbus_proxy, &error)) {
		g_error ("Error invoking org_gnome_Rhythmbox_get_playing_time: %s", error->message);
		g_error_free (error);
	}
}

static long
rb_remote_dbus_client_get_playing_time_impl (RBRemoteClientProxy *proxy)
{
	RBRemoteDBus *dbus = RB_REMOTE_DBUS (proxy);
	guint32 playing_time;
	GError *error = NULL;
	
	if (dbus->priv->dbus_proxy == NULL)
		rb_remote_dbus_create_client_proxy (dbus);
	
	if (!org_gnome_Rhythmbox_get_playing_time (dbus->priv->dbus_proxy, &playing_time, &error)) {
		g_error ("Error invoking org_gnome_Rhythmbox_get_playing_time: %s", error->message);
		g_error_free (error);
		return -1;
	}

	return (long)playing_time;
}

static void
rb_remote_dbus_client_set_playing_time_impl (RBRemoteClientProxy *proxy, long time)
{
	RBRemoteDBus *dbus = RB_REMOTE_DBUS (proxy);
	GError *error = NULL;
	
	if (dbus->priv->dbus_proxy == NULL)
		rb_remote_dbus_create_client_proxy (dbus);

	if (!org_gnome_Rhythmbox_set_playing_time (dbus->priv->dbus_proxy, time, &error)) {
		g_error ("Error invoking org_gnome_Rhythmbox_set_playing_time: %s", error->message);
		g_error_free (error);
	}
}

static void
rb_remote_dbus_client_quit_impl (RBRemoteClientProxy *proxy)
{
	RBRemoteDBus *dbus = RB_REMOTE_DBUS (proxy);
	GError *error = NULL;
	
	if (dbus->priv->dbus_proxy == NULL)
		rb_remote_dbus_create_client_proxy (dbus);


	if (!org_gnome_Rhythmbox_quit (dbus->priv->dbus_proxy, &error)) {
		g_error ("Error invoking org_gnome_Rhythmbox_quit: %s", error->message);
		g_error_free (error);
	}
}
