/*
 *  arch-tag: minimal DBus client program
 *
 *  Copyright (C) 2005 Jonathan Matthew <jonathan@kaolin.hn.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>
#include <termios.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <assert.h>
#include <popt.h>
#include <string.h>

#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>
#include <dbus/dbus-glib-bindings.h>

#include "org_gnome_Rhythmbox-binding.h"

#define RB_REMOTE_DBUS_SERVICE_PATH            "org.gnome.Rhythmbox"
#define RB_REMOTE_DBUS_NAME "org.gnome.Rhythmbox"
#define RB_REMOTE_DBUS_OBJECT_PATH             "/org/gnome/Rhythmbox/DBusRemote"

static int print_playing_title  = FALSE;
static int print_playing_artist = FALSE;
static int print_playing_album  = FALSE;
static int print_playing_track  = FALSE;
static int print_playing_disc   = FALSE;
static int print_playing_genre  = FALSE;
static int print_playing_path   = FALSE;
static int print_shuffle        = FALSE;
static int print_repeat         = FALSE;
static int playpause            = FALSE;
static int do_play              = FALSE;
static int do_pause             = FALSE;
static int focus                = FALSE;
static int previous             = FALSE;
static int next                 = FALSE;
static int shuffle              = FALSE;
static int repeat               = FALSE;
static int print_play_time      = FALSE;
static int print_song_length    = FALSE;
static int quit			= FALSE;
static int notify               = FALSE;
static long seek_time           = -1;
static int set_shuffle          = -1;
static int set_repeat           = -1;

static void
bail_on_error (const char *what, GError *error)
{
	if (error != NULL) {
		printf ("%s: %s\n", what, error->message);
		g_error_free (error);
		exit (1);	/* ugh */
	}
}

static void
shuffle_changed_cb (DBusGProxy *proxy, gboolean shuffle, gpointer blah)
{
	printf ("- shuffle changed; now %s\n", shuffle ? "on" : "off");
}

static void
repeat_changed_cb (DBusGProxy *proxy, gboolean repeat, gpointer blah)
{
	printf ("- repeat changed; now %s\n", repeat ? "on" : "off");
}

int
main (int argc, const char **argv)
{
        DBusGConnection  *conn;
	DBusGProxy	 *proxy;
        GError           *error = NULL;
        int               rc;
        poptContext       popt_ctx;

        struct poptOption popt_options [] = {
		{ "print-playing-title",  0, POPT_ARG_NONE, &print_playing_title,  0, "Print the playing song title", NULL },
		{ "print-playing-artist", 0, POPT_ARG_NONE, &print_playing_artist, 0, "Print the playing song artist", NULL },
		{ "print-playing-album",  0, POPT_ARG_NONE, &print_playing_album,  0, "Print the playing song album", NULL },
		{ "print-playing-track",  0, POPT_ARG_NONE, &print_playing_track,  0, "Print the playing song track number", NULL },
		{ "print-playing-disc",   0, POPT_ARG_NONE, &print_playing_disc,   0, "Print the playing song disc number", NULL },
		{ "print-playing-genre",  0, POPT_ARG_NONE, &print_playing_genre,  0, "Print the playing song genre", NULL },
		{ "print-playing-path",	  0, POPT_ARG_NONE, &print_playing_path,   0, "Print the playing song URI", NULL },
        
                { "print-song-length",    0, POPT_ARG_NONE, &print_song_length,    0, "Print the playing song length in seconds", NULL },
                { "print-play-time",      0, POPT_ARG_NONE, &print_play_time,      0, "Print the current elapsed time of playing song", NULL },
                { "set-play-time",        0, POPT_ARG_LONG, &seek_time,            0, "Seek to the specified time in playing song if possible", NULL },
        
		{ "play-pause",           0, POPT_ARG_NONE, &playpause,            0, "Toggle play/pause mode", NULL },
                { "play",                 0, POPT_ARG_NONE, &do_play,              0, "Start playback", NULL },
                { "pause",                0, POPT_ARG_NONE, &do_pause,             0, "Pause playback", NULL},

		{ "focus",                0, POPT_ARG_NONE, &focus,                0, "Focus the running player", NULL },
		{ "previous",             0, POPT_ARG_NONE, &previous,             0, "Jump to previous song", NULL },
		{ "next",                 0, POPT_ARG_NONE, &next,                 0, "Jump to next song", NULL },
		
		{ "shuffle",              0, POPT_ARG_NONE, &shuffle,              0, "Toggle shuffling", NULL },
		{ "repeat",               0, POPT_ARG_NONE, &repeat,               0, "Toggle repeat", NULL },

                { "set-shuffle",          0, POPT_ARG_INT,  &set_shuffle,          0, "Set shuffle on or off", NULL },
                { "set-repeat",           0, POPT_ARG_INT,  &set_repeat,           0, "Set repeat on or off", NULL },
                { "print-shuffle",        0, POPT_ARG_NONE, &print_shuffle,        0, "Print shuffle status", NULL },
                { "print-repeat",         0, POPT_ARG_NONE, &print_repeat,         0, "Print repeat status", NULL },

		{ "quit",                 0, POPT_ARG_NONE, &quit,                 0, "Quit rhythmbox", NULL },

		{ "notify",               0, POPT_ARG_NONE, &notify,               0, "Wait for and display notifications from rhythmbox", NULL },

		POPT_TABLEEND
        };

	g_type_init ();
	
        popt_ctx = poptGetContext (NULL, argc, argv, popt_options, 0);
        
        if (argc < 2) {
                poptPrintUsage (popt_ctx, stderr, 0);
                return 1;
        }

        rc = poptGetNextOpt (popt_ctx);

        if (rc != -1) {
                poptPrintUsage (popt_ctx, stderr, 0);
                return 1;
        }

        /* Acquire DBus connection */
	conn = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
	bail_on_error ("Couldn't connect to session bus", error);

	/* Create proxy for rhythmbox remote object */
	proxy = dbus_g_proxy_new_for_name_owner (conn,
					   RB_REMOTE_DBUS_NAME,
					   RB_REMOTE_DBUS_OBJECT_PATH,
					   RB_REMOTE_DBUS_SERVICE_PATH,
					   &error);
	bail_on_error ("Couldn't find Rhythmbox remote control", error);

        /* Execute user commands */
        if (focus) {
		org_gnome_Rhythmbox_grab_focus (proxy, &error);
		bail_on_error ("Grabbing focus", error);
        }

        if (shuffle) {
		org_gnome_Rhythmbox_toggle_shuffle (proxy, &error);
		bail_on_error ("Toggling shuffle", error);
        }

        if (repeat) {
		org_gnome_Rhythmbox_toggle_repeat (proxy, &error);
		bail_on_error ("Toggling repeat", error);
        }

        if (playpause) {
		org_gnome_Rhythmbox_play_pause (proxy, &error);
		bail_on_error ("Playing/pausing", error);
        }

        if (do_play) {
		org_gnome_Rhythmbox_play (proxy, &error);
		bail_on_error ("Starting playback", error);
        }

        if (do_pause) {
		org_gnome_Rhythmbox_pause (proxy, &error);
		bail_on_error ("Pausing playback", error);
        }

        if (next) {
		org_gnome_Rhythmbox_next (proxy, &error);
		bail_on_error ("Jumping to next song", error);
        }

        if (previous) {
		org_gnome_Rhythmbox_previous (proxy, &error);
		bail_on_error ("Jumping to previous song", error);
        }

        if (set_shuffle != -1) {
		org_gnome_Rhythmbox_set_shuffle (proxy, set_shuffle, &error);
		bail_on_error ("Setting shuffle", error);
        }

        if (set_repeat != -1) {
		org_gnome_Rhythmbox_set_repeat (proxy, set_repeat, &error);
		bail_on_error ("Setting repeat", error);
        }

        if (print_repeat) {
		gboolean repeat;
		org_gnome_Rhythmbox_get_repeat (proxy, &repeat, &error);
		bail_on_error ("Getting repeat setting", error);
		printf ("%s\n", repeat ? "repeat" : "no repeat");
        }

        if (print_shuffle) {
		gboolean shuffle;
		org_gnome_Rhythmbox_get_shuffle (proxy, &shuffle, &error);
		bail_on_error ("Getting shuffle setting", error);
		printf ("%s\n", shuffle ? "shuffle" : "no shuffle");
        }

        if (print_play_time) {
		guint32 time;
		org_gnome_Rhythmbox_get_playing_time (proxy, &time, &error);
		bail_on_error ("Getting playing time", error);
                printf ("%u:%.2u\n", time/60, time % 60);
        }

        if (seek_time >= 0) {
		org_gnome_Rhythmbox_set_playing_time (proxy, seek_time, &error);
		bail_on_error ("Setting playback position", error);
        }

        if (print_playing_artist || print_playing_album ||
            print_playing_track  || print_playing_genre ||
            print_playing_path   || print_song_length) {

		/* hmm, ugly */
		char *uri;
		char *title;
		char *artist;
		char *genre;
		char *album;
		guint32 track_number;
		guint32 duration;
		guint32 bitrate;
		guint32 filesize;
		guint32 rating;
		guint32 play_count;
		guint32 last_played;
		guint32 disc_number;

		org_gnome_Rhythmbox_get_playing_song_ur_i (proxy, &uri, &error);
		bail_on_error ("Getting playing song URI", error);
		
		org_gnome_Rhythmbox_get_song_info (proxy, uri, &title, &artist,
						   &genre, &album, &track_number,
						   &duration, &bitrate, &filesize,
						   &rating, &play_count, &last_played,
						   &disc_number, &error);
		bail_on_error ("Getting song information", error);

		if (print_playing_title && title)
			printf ("Title: %s\n", title);

		if (print_playing_artist && artist)
			printf ("Artist: %s\n", artist);

		if (print_playing_album && album)
			printf ("Album: %s\n", album);

		if (print_playing_genre && genre)
			printf ("Genre: %s\n", genre);

		if (print_playing_path && uri)
			printf ("URI: %s\n", uri);

		if (print_playing_track)
			printf ("Track number: %u\n", track_number);

		if (print_playing_disc)
			printf ("Disc number: %u\n", disc_number);

		if (print_song_length)
			printf ("Length: %u:%.2u\n", duration / 60, duration % 60);

		g_free (uri);
		g_free (title);
		g_free (artist);
		g_free (genre);
		g_free (album);
        }

	if (quit) {
		org_gnome_Rhythmbox_quit (proxy, &error);
		bail_on_error ("Making Rhythmbox quit", error);
	} else if (notify) {
		/* well, you're not going to get any if you just told it to quit */

		GMainLoop *mainloop;

		dbus_g_proxy_add_signal (proxy, "ShuffleChanged", "b");
		dbus_g_proxy_add_signal (proxy, "RepeatChanged", "b");
		dbus_g_proxy_connect_signal (proxy, "ShuffleChanged",
					     G_CALLBACK (shuffle_changed_cb),
					     NULL, NULL);
		dbus_g_proxy_connect_signal (proxy, "RepeatChanged",
					     G_CALLBACK (repeat_changed_cb),
					     NULL, NULL);


		mainloop = g_main_loop_new (NULL, FALSE);
		g_main_loop_run (mainloop);
	}

        return 0;
}

