/* 
 *  arch-tag: Implementation of queue panel toggle container
 *
 *  Copyright (C) 2003 Jorn Baayen <jorn@nl.linux.org>
 *  Copyright (C) 2003,2004 Colin Walters <walters@redhat.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <gtk/gtk.h>
#include <config.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <libgnome/gnome-i18n.h>

#include "disclosure-widget.h"
#include "rb-queue-header.h"
#include "rb-stock-icons.h"
#include "rb-preferences.h"
#include "rb-debug.h"
#include "rb-remote.h"
#include "eel-gconf-extensions.h"
#include "rb-preferences.h"
#include "rb-playlist-source.h"

static const GtkTargetEntry rb_queue_header_drag_types[] = {{  "text/uri-list", 0, 0 }};

static void rb_queue_header_class_init (RBQueueHeaderClass *klass);
static void rb_queue_header_init (RBQueueHeader *queue_header);
static void rb_queue_header_finalize (GObject *object);
static void rb_queue_header_set_property (GObject *object,
					  guint prop_id,
					  const GValue *value,
					  GParamSpec *pspec);
static void rb_queue_header_get_property (GObject *object,
					  guint prop_id,
					  GValue *value,
					  GParamSpec *pspec);

static void rb_queue_header_disclosure_toggled_cb (GtkToggleButton *disclosure,
						    gpointer data);
static void rb_queue_header_set_disclosure_labels (RBQueueHeader *queue_header);
static void rb_queue_header_queue_row_inserted_cb (GtkTreeModel *treemodel, 
						   GtkTreePath *arg1, 
						   GtkTreeIter *arg2, 
						   gpointer data);
static void rb_queue_header_queue_row_deleted_cb (GtkTreeModel *treemodel, 
						  GtkTreePath *arg1, 
						  gpointer data);
static void rb_queue_header_drag_data_received_cb (GtkWidget *widget,
						   GdkDragContext *dc,
						   gint x,
						   gint y,
						   GtkSelectionData *data,
						   guint info,
						   guint t,
						   RBQueueHeader *header);

static guint get_queue_length (RBQueueHeader *header);
static gboolean _idle_set_disclosure_labels (gpointer header);

struct RBQueueHeaderPrivate
{
	GtkActionGroup *actiongroup;
	GtkWidget *disclosure;

	guint queue_notify_id;

	RBSource *selected_source;
	RBPlaylistSource *queue_source;
};

enum
{
	PROP_0,
	PROP_ACTION_GROUP,
	PROP_SOURCE,
	PROP_QUEUE_SOURCE,
};

static GObjectClass *parent_class = NULL;

GType
rb_queue_header_get_type (void)
{
	static GType rb_queue_header_type = 0;

	if (rb_queue_header_type == 0)
	{
		static const GTypeInfo our_info =
		{
			sizeof (RBQueueHeaderClass),
			NULL,
			NULL,
			(GClassInitFunc) rb_queue_header_class_init,
			NULL,
			NULL,
			sizeof (RBQueueHeader),
			0,
			(GInstanceInitFunc) rb_queue_header_init
		};

		rb_queue_header_type = g_type_register_static (GTK_TYPE_HBOX,
							       "RBQueueHeader",
							       &our_info, 0);
	}

	return rb_queue_header_type;
}

static void
rb_queue_header_class_init (RBQueueHeaderClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

	parent_class = g_type_class_peek_parent (klass);

	object_class->finalize = rb_queue_header_finalize;

	object_class->set_property = rb_queue_header_set_property;
	object_class->get_property = rb_queue_header_get_property;

	g_object_class_install_property (object_class,
					 PROP_ACTION_GROUP,
					 g_param_spec_object ("action-group",
							      "GtkActionGroup",
							      "GtkActionGroup object",
							      GTK_TYPE_ACTION_GROUP,
							      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
	g_object_class_install_property (object_class,
					 PROP_SOURCE,
					 g_param_spec_object ("source",
						 	      "RBSource",
							      "RBSource object",
							      RB_TYPE_SOURCE,
							      G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_QUEUE_SOURCE,
					 g_param_spec_object ("queue-source",
						 	      "RBPlaylistSource",
							      "RBPlaylistSource object",
							      RB_TYPE_PLAYLIST_SOURCE,
							      G_PARAM_READWRITE));
}

static void
rb_queue_header_init (RBQueueHeader *header)
{
	header->priv = g_new0 (RBQueueHeaderPrivate, 1);

	header->priv->disclosure = cddb_disclosure_new ("", "");	/* labels filled in later */
	gtk_widget_set_sensitive (header->priv->disclosure, TRUE);
	rb_queue_header_set_disclosure_labels (header);
	g_signal_connect_object (G_OBJECT (header->priv->disclosure), "toggled",
				 G_CALLBACK (rb_queue_header_disclosure_toggled_cb), header, 0);

	gtk_box_pack_start (GTK_BOX (header), header->priv->disclosure, TRUE, TRUE, 0);

	g_signal_connect (G_OBJECT (header->priv->disclosure),
			  "drag_data_received",
			  G_CALLBACK (rb_queue_header_drag_data_received_cb),
			  header);


}

static void
rb_queue_header_finalize (GObject *object)
{
	RBQueueHeader *header;

	g_return_if_fail (object != NULL);
	g_return_if_fail (RB_IS_QUEUE_HEADER (object));

	header = RB_QUEUE_HEADER (object);

	g_return_if_fail (header->priv != NULL);

	g_free (header->priv);

	G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
rb_queue_header_set_property (GObject *object,
			      guint prop_id,
			      const GValue *value,
			      GParamSpec *pspec)
{
	RBQueueHeader *header = RB_QUEUE_HEADER (object);

	switch (prop_id)
	{
	case PROP_ACTION_GROUP:
		header->priv->actiongroup = g_value_get_object (value);
		break;
	case PROP_SOURCE:
		header->priv->selected_source = g_value_get_object (value);
		rb_queue_header_sync_control_state_int (header);
		break;
	case PROP_QUEUE_SOURCE:
		if (header->priv->queue_source) {
			RhythmDBQueryModel *model = rb_playlist_source_get_model (header->priv->queue_source);
			g_signal_handlers_disconnect_by_func (G_OBJECT (model),
							      G_CALLBACK (rb_queue_header_queue_row_inserted_cb),
							      header);
			g_signal_handlers_disconnect_by_func (G_OBJECT (model),
							      G_CALLBACK (rb_queue_header_queue_row_deleted_cb),
							      header);

			gtk_drag_dest_unset (GTK_WIDGET (header->priv->disclosure));
		}

		header->priv->queue_source = g_value_get_object (value);
		rb_queue_header_set_disclosure_labels (header);

		if (header->priv->queue_source) {
			RhythmDBQueryModel *model = rb_playlist_source_get_model (header->priv->queue_source);

			g_signal_connect_object (G_OBJECT (model),
						 "row-inserted", 
						 G_CALLBACK (rb_queue_header_queue_row_inserted_cb), 
						 header, 0);
			g_signal_connect_object (G_OBJECT (model),
						 "row-deleted", 
						 G_CALLBACK (rb_queue_header_queue_row_deleted_cb), 
						 header, 0);

			gtk_drag_dest_set (GTK_WIDGET (header->priv->disclosure),
					   GTK_DEST_DEFAULT_ALL,
					   rb_queue_header_drag_types, 1,
					   GDK_ACTION_COPY | GDK_ACTION_MOVE);
		}

		rb_queue_header_sync_control_state_int (header);

		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

void
rb_queue_header_set_source (RBQueueHeader *header,
			    RBSource *source)
{
	g_return_if_fail (RB_IS_QUEUE_HEADER (header));
	g_return_if_fail (RB_IS_SOURCE (source));

	g_object_set (G_OBJECT (header), "source", source, NULL);
}

static void 
rb_queue_header_get_property (GObject *object,
			      guint prop_id,
			      GValue *value,
			      GParamSpec *pspec)
{
	RBQueueHeader *header = RB_QUEUE_HEADER (object);

	switch (prop_id)
	{
	case PROP_ACTION_GROUP:
		g_value_set_object (value, header->priv->actiongroup);
		break;
	case PROP_QUEUE_SOURCE:
		g_value_set_object (value, header->priv->queue_source);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

RBQueueHeader *
rb_queue_header_new (GtkActionGroup *actiongroup)
{
	RBQueueHeader *header = g_object_new (RB_TYPE_QUEUE_HEADER,
					       "action-group", actiongroup,
					       NULL);

	g_return_val_if_fail (header->priv != NULL, NULL);

	return header;
}

static void
rb_queue_header_disclosure_toggled_cb (GtkToggleButton *disclosure,
					gpointer data)
{
	gboolean disclosed = gtk_toggle_button_get_active (disclosure);

	rb_debug ("disclosed: %s", disclosed ? "TRUE" : "FALSE");
	
	eel_gconf_set_boolean (CONF_UI_QUEUE_HIDDEN, !disclosed);
}

void
rb_queue_header_sync_control_state_int (RBQueueHeader *header)
{
	rb_queue_header_sync_control_state (header->priv->disclosure, 
					    GTK_WIDGET (header),
					    header->priv->selected_source,
					    RB_SOURCE (header->priv->queue_source),
					    header->priv->actiongroup);
}

void
rb_queue_header_sync_control_state (GtkWidget *disclosure, GtkWidget *header, RBSource *selected_source, RBSource *queue_source, GtkActionGroup *actiongroup)
{
	gboolean have_queue = TRUE;
	gboolean not_small = !eel_gconf_get_boolean (CONF_UI_SMALL_DISPLAY);
	gboolean hidden = eel_gconf_get_boolean (CONF_UI_QUEUE_HIDDEN);

	if (selected_source != NULL)
		have_queue = rb_source_should_show_queue (selected_source);
	
	rb_debug ("syncing queue header control state; have queue: %d, hidden: %d, not small: %d", have_queue, hidden, not_small);

	if (header != NULL) {
		if (have_queue)
			gtk_widget_show (GTK_WIDGET (header));
		else
			gtk_widget_hide (GTK_WIDGET (header));
	}

#if !defined(MAYBE_A_BIT_UGLY)
	if (actiongroup != NULL) {
		GtkAction *viewqueue_action;
		viewqueue_action = gtk_action_group_get_action (actiongroup,
								"ViewQueue");
		g_object_set (G_OBJECT (viewqueue_action), "sensitive",
			      have_queue && not_small, NULL);
		gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (viewqueue_action),
					      !hidden);
	}

	if (queue_source != NULL) {
		if (have_queue && !hidden)
			gtk_widget_show (GTK_WIDGET (queue_source));
		else
			gtk_widget_hide (GTK_WIDGET (queue_source));
	}
#endif

	if (disclosure != NULL)
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (disclosure), !hidden);
}

static guint
get_queue_length (RBQueueHeader *header)
{
	if (header->priv->queue_source != NULL) {
		RhythmDBQueryModel *model = rb_playlist_source_get_model (header->priv->queue_source);
		return gtk_tree_model_iter_n_children (GTK_TREE_MODEL (model), NULL);
	}

	return 0;
}

static void
rb_queue_header_queue_row_inserted_cb (GtkTreeModel *treemodel, 
				       GtkTreePath *arg1, 
				       GtkTreeIter *arg2,
				       gpointer data)
{
	RBQueueHeader *header;

	rb_debug ("Row inserted into queue model");

	header = RB_QUEUE_HEADER (data);
	rb_queue_header_set_disclosure_labels (header);

#if !defined(MAYBE_A_BIT_UGLY)
	/* if the queue was empty beforehand, expand it */
	if (get_queue_length (header) == 1)
		eel_gconf_set_boolean (CONF_UI_QUEUE_HIDDEN, FALSE);
#endif
}

static void
rb_queue_header_queue_row_deleted_cb (GtkTreeModel *treemodel, 
				      GtkTreePath *arg1, 
				      gpointer data)
{
	RBQueueHeader *header;

	rb_debug ("Row removed from queue model");

	header = RB_QUEUE_HEADER (data);
	/* this is called before the row is removed, so we can't update the labels yet */
	g_idle_add (_idle_set_disclosure_labels, header);

#if !defined(MAYBE_A_BIT_UGLY)
	/* if the queue is now empty, collapse it */
	if (get_queue_length (header) == 1)
		eel_gconf_set_boolean (CONF_UI_QUEUE_HIDDEN, TRUE);
#endif
}

static gboolean
_idle_set_disclosure_labels (gpointer header)
{
	rb_queue_header_set_disclosure_labels (header);
	return FALSE;
}

static void
rb_queue_header_set_disclosure_labels (RBQueueHeader *header)
{
	gchar *hidden_label;
	gchar *queue_size_string;

	if (header->priv->disclosure == NULL)
		return;

	if (get_queue_length (header) > 0)
		queue_size_string = rb_source_get_status (RB_SOURCE (header->priv->queue_source));
	else
		queue_size_string = g_strdup (_("empty"));

	hidden_label = g_strdup_printf (_("Show Queue (%s)"), queue_size_string);
	cddb_disclosure_set_labels (CDDB_DISCLOSURE (header->priv->disclosure), 
				    _("Hide Queue"), 
				    hidden_label);
	g_free (hidden_label);
	g_free (queue_size_string);

	rb_queue_header_sync_control_state_int (header);
}

static void
rb_queue_header_drag_data_received_cb (GtkWidget *widget,
				       GdkDragContext *context,
				       gint x,
				       gint y,
				       GtkSelectionData *selection_data,
				       guint info,
				       guint time,
				       RBQueueHeader *header)
{
	rb_debug ("received dropped data on the queue disclosure widget");

	/* pass the drop on to the queue source */
	rb_source_receive_drag (RB_SOURCE (header->priv->queue_source), selection_data);
}

