/*
 *  arch-tag: Implementation of RhythmDB query creation dialog
 *
 *  Copyright (C) 2003, 2004 Colin Walters <walters@gnome.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <config.h>
#include <string.h>
#include <stdlib.h>
#include <libgnome/gnome-i18n.h>
#include <glib/gprintf.h>
#include <gtk/gtk.h>

#include "rhythmdb.h"
#include "rb-query-creator.h"
#include "rb-dialog.h"
#include "rb-rating.h"
#include "rb-debug.h"
#include "rb-preferences.h"
#include "rb-glade-helpers.h"


typedef struct
{
	const char *name;
	gboolean strict;
	RhythmDBPropType val;
} RBQueryCreatorCriteriaOption;

typedef enum {
	RHYTHM_QUERY_ENTRY_TEXT,
	RHYTHM_QUERY_ENTRY_RATING,
	RHYTHM_QUERY_ENTRY_POSITIVE_INTEGER,
	RHYTHM_QUERY_ENTRY_DURATION,
	RHYTHM_QUERY_ENTRY_RELATIVE_TIME,
} RBQueryCreatorEntryType;

typedef struct
{
	const char *name;
	RhythmDBPropType strict_val;
	RhythmDBPropType fuzzy_val;;
	const RBQueryCreatorCriteriaOption *criteria_options;
	guint criteria_options_length;
	RBQueryCreatorEntryType entry_type;
} RBQueryCreatorPropertyOption;

typedef struct
{
	const char *name;
	gulong timeMultiplier;
} RBQueryCreatorTimeUnitOption;

typedef struct
{
	const char *name;
	const char *sort_order;
} RBQueryCreatorSortOption;

const RBQueryCreatorCriteriaOption string_criteria_options[] =
{
	{ N_("contains"), 0, RHYTHMDB_QUERY_PROP_LIKE },
	{ N_("does not contain"), 0, RHYTHMDB_QUERY_PROP_NOT_LIKE },
	{ N_("equals"), 1, RHYTHMDB_QUERY_PROP_EQUALS }
};

const RBQueryCreatorCriteriaOption numeric_criteria_options[] =
{
	{ N_("equals"), 1, RHYTHMDB_QUERY_PROP_EQUALS },
	{ N_("greater than"), 1, RHYTHMDB_QUERY_PROP_GREATER },
	{ N_("less than"), 1, RHYTHMDB_QUERY_PROP_LESS },
	{ N_("at least"), 1, RHYTHMDB_QUERY_PROP_GREATER_EQUALS },
	{ N_("at most"), 1, RHYTHMDB_QUERY_PROP_LESS_EQUALS }
};

const RBQueryCreatorCriteriaOption time_criteria_options[] =
{
	{ N_("in the last"), 1, RHYTHMDB_QUERY_PROP_CURRENT_TIME_WITHIN },
	{ N_("not in the last"), 1, RHYTHMDB_QUERY_PROP_CURRENT_TIME_NOT_WITHIN }
};

const RBQueryCreatorTimeUnitOption time_unit_options[] =
{
	{ N_("seconds"), 1 },
	{ N_("minutes"), 60 },
	{ N_("hours"), 60 * 60 },
	{ N_("days"), 60 * 60 * 24 },
	{ N_("weeks"), 60 * 60 * 24 * 7 }
};

/* set the default time unit to be days*/
const gulong time_unit_options_default = 3;

/*
 * These need to match up with the column keys from rb-entry-view.c
 */
const RBQueryCreatorSortOption sort_options[] =
{
	{ N_("Title"), "Title" },
	{ N_("Artist"), "Artist" },
	{ N_("Album"), "Album" },
	{ N_("Genre"), "Genre" },
	{ N_("Rating"), "Rating" },
	{ N_("Play Count"), "PlayCount" },
	{ N_("Track Number"), "Track"},
	{ N_("Duration"), "Time" },
	{ N_("Last Played"), "LastPlayed" }
};

const RBQueryCreatorPropertyOption property_options[] =
{
	{ N_("Title"), RHYTHMDB_PROP_TITLE, RHYTHMDB_PROP_TITLE_FOLDED, string_criteria_options, G_N_ELEMENTS (string_criteria_options), RHYTHM_QUERY_ENTRY_TEXT },
	{ N_("Artist"), RHYTHMDB_PROP_ARTIST, RHYTHMDB_PROP_ARTIST_FOLDED, string_criteria_options, G_N_ELEMENTS (string_criteria_options), RHYTHM_QUERY_ENTRY_TEXT },
	{ N_("Album"), RHYTHMDB_PROP_ALBUM, RHYTHMDB_PROP_ALBUM_FOLDED, string_criteria_options, G_N_ELEMENTS (string_criteria_options), RHYTHM_QUERY_ENTRY_TEXT },
	{ N_("Genre"), RHYTHMDB_PROP_GENRE, RHYTHMDB_PROP_GENRE_FOLDED, string_criteria_options, G_N_ELEMENTS (string_criteria_options), RHYTHM_QUERY_ENTRY_TEXT },
	{ N_("Rating"), RHYTHMDB_PROP_RATING, RHYTHMDB_PROP_RATING, numeric_criteria_options, G_N_ELEMENTS (numeric_criteria_options), RHYTHM_QUERY_ENTRY_RATING },
	{ N_("Play Count"), RHYTHMDB_PROP_PLAY_COUNT, RHYTHMDB_PROP_PLAY_COUNT, numeric_criteria_options, G_N_ELEMENTS (numeric_criteria_options), RHYTHM_QUERY_ENTRY_POSITIVE_INTEGER },
	{ N_("Track Number"), RHYTHMDB_PROP_TRACK_NUMBER, RHYTHMDB_PROP_TRACK_NUMBER, numeric_criteria_options, G_N_ELEMENTS (numeric_criteria_options), RHYTHM_QUERY_ENTRY_POSITIVE_INTEGER },
	{ N_("Disc Number"), RHYTHMDB_PROP_DISC_NUMBER, RHYTHMDB_PROP_DISC_NUMBER, numeric_criteria_options, G_N_ELEMENTS (numeric_criteria_options), RHYTHM_QUERY_ENTRY_POSITIVE_INTEGER },
	{ N_("Bitrate"), RHYTHMDB_PROP_BITRATE, RHYTHMDB_PROP_BITRATE, numeric_criteria_options, G_N_ELEMENTS (numeric_criteria_options), RHYTHM_QUERY_ENTRY_POSITIVE_INTEGER },
	{ N_("Duration"), RHYTHMDB_PROP_DURATION, RHYTHMDB_PROP_DURATION, numeric_criteria_options, G_N_ELEMENTS (numeric_criteria_options), RHYTHM_QUERY_ENTRY_DURATION },
	{ N_("First Seen"), RHYTHMDB_PROP_FIRST_SEEN, RHYTHMDB_PROP_FIRST_SEEN, time_criteria_options, G_N_ELEMENTS (time_criteria_options), RHYTHM_QUERY_ENTRY_RELATIVE_TIME },
	{ N_("Last Played"), RHYTHMDB_PROP_LAST_PLAYED, RHYTHMDB_PROP_LAST_PLAYED, time_criteria_options, G_N_ELEMENTS (time_criteria_options), RHYTHM_QUERY_ENTRY_RELATIVE_TIME }
	
	/* properties not included
	
	RHYTHMDB_PROP_MTIME,		(gulong)
	RHYTHMDB_PROP_LAST_SEEN,	(gulong)
	RHYTHMDB_PROP_RELEASE_DATE,	(GDate) [not int main branch],
	
	RHYTHMDB_PROP_FILE_SIZE,	(guint64)
	RHYTHMDB_PROP_LOCATION,		(char *)
	RHYTHMDB_PROP_MOUNTPOINT,	(char *)
	RHYTHMDB_PROP_AUTO_RATE,	(gboolean)
	RHYTHMDB_PROP_TRACK_GAIN,	(double)
	RHYTHMDB_PROP_TRACK_PEAK,	(double)
	RHYTHMDB_PROP_ALBUM_GAIN,	(double)
	RHYTHMDB_PROP_ALBUM_PEAK,	(double)
	RHYTHMDB_PROP_MIMETYPE,		(RBRefString)
	*/
};

const char * DEFAULT_SORTING_COLUMN = "Title";
const gint DEFAULT_SORTING_ORDER = GTK_SORT_ASCENDING;




static void rb_query_creator_class_init (RBQueryCreatorClass *klass);
static void rb_query_creator_init (RBQueryCreator *dlg);
static GObject *rb_query_creator_constructor (GType type, guint n_construct_properties,
					      GObjectConstructParam *construct_properties);
static void rb_query_creator_finalize (GObject *object);
static void rb_query_creator_set_property (GObject *object,
					      guint prop_id,
					      const GValue *value,
					      GParamSpec *pspec);
static void rb_query_creator_get_property (GObject *object,
					      guint prop_id,
					      GValue *value,
					      GParamSpec *pspec);
static gboolean select_criteria_from_value (RBQueryCreator *creator,
					GtkWidget *option_menu,
					RhythmDBPropType prop,
					RhythmDBQueryType qtype);

static void setup_property_option_menu (RBQueryCreator *creator,
					GtkWidget *option_menu,
					const RBQueryCreatorPropertyOption *options,
					int length, gboolean activate_first);
static void setup_criteria_option_menu (GtkWidget *option_menu,
					const RBQueryCreatorCriteriaOption *options,
					int length, gboolean activate_first);
static void setup_time_unit_option_menu (GtkWidget *option_menu,
					const RBQueryCreatorTimeUnitOption *options,
					int length, gboolean activate_first);
static void setup_sort_option_menu (GtkWidget *option_menu,
			    const RBQueryCreatorSortOption *options,
			    int length, gboolean activate_first);
static GtkWidget * create_property_option_menu (RBQueryCreator *creator,
						const RBQueryCreatorPropertyOption *options,
						int length, gboolean activate_first);
static GtkWidget * create_criteria_option_menu (const RBQueryCreatorCriteriaOption *options,
						int length, gboolean activate_first);
static GtkWidget* create_time_unit_option_menu (const RBQueryCreatorTimeUnitOption *options,
			     		int length, gboolean activate_first);
static GtkWidget * option_menu_get_active_child (GtkWidget *option_menu);
static GtkWidget * get_box_widget_at_pos (GtkBox *box, guint pos);

static GtkWidget * append_row (RBQueryCreator *dialog);
static void add_button_click_cb (GtkWidget *button, RBQueryCreator *creator);
static void remove_button_click_cb (GtkWidget *button, RBQueryCreator *creator);
static void limit_toggled_cb (GtkWidget *limit, RBQueryCreator *creator);
static gint get_property_options_index(RhythmDBPropType prop);

struct RBQueryCreatorPrivate
{
	RhythmDB *db;

	gboolean creating;
	
	GtkSizeGroup *property_size_group;
	GtkSizeGroup *criteria_size_group;
	GtkSizeGroup *entry_size_group;
	GtkSizeGroup *button_size_group;

	GtkBox *vbox;
	GList *rows;

	GtkWidget *addbutton;
	GtkWidget *disjunction_check;
	GtkWidget *limit_check;
	GtkWidget *limit_entry;
	GtkWidget *limit_option;
	GtkWidget *sort_reverse_check;
	GtkWidget *sort_option;
};

static GObjectClass *parent_class = NULL;

enum
{
	PROP_0,
	PROP_DB,
	PROP_CREATING,
};

GType
rb_query_creator_get_type (void)
{
	static GType rb_query_creator_type = 0;

	if (rb_query_creator_type == 0)
	{
		static const GTypeInfo our_info =
		{
			sizeof (RBQueryCreatorClass),
			NULL,
			NULL,
			(GClassInitFunc) rb_query_creator_class_init,
			NULL,
			NULL,
			sizeof (RBQueryCreator),
			0,
			(GInstanceInitFunc) rb_query_creator_init
		};
		
		rb_query_creator_type = g_type_register_static (GTK_TYPE_DIALOG,
								"RBQueryCreator",
								&our_info, 0);
	}

	return rb_query_creator_type;
}

static void
rb_query_creator_class_init (RBQueryCreatorClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

	parent_class = g_type_class_peek_parent (klass);

	object_class->finalize = rb_query_creator_finalize;
	object_class->constructor = rb_query_creator_constructor;
	object_class->set_property = rb_query_creator_set_property;
	object_class->get_property = rb_query_creator_get_property;

	g_object_class_install_property (object_class,
					 PROP_DB,
					 g_param_spec_object ("db",
							      "RhythmDB",
							      "RhythmDB database",
							      RHYTHMDB_TYPE,
							      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));

	g_object_class_install_property (object_class,
					 PROP_CREATING,
					 g_param_spec_boolean ("creating",
							       "creating",
							       "Whether or not we're creating a new playlist",
							       TRUE,
							       G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
}

static GObject *
rb_query_creator_constructor (GType type, guint n_construct_properties,
			      GObjectConstructParam *construct_properties)
{
	RBQueryCreator *dlg;
	RBQueryCreatorClass *klass;
	GObjectClass *parent_class;  
	GladeXML *xml;
	GtkWidget *mainbox;

	klass = RB_QUERY_CREATOR_CLASS (g_type_class_peek (type));
	parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
	dlg = RB_QUERY_CREATOR (parent_class->constructor (type, n_construct_properties,
							   construct_properties));
	
	if (dlg->priv->creating) { 
		gtk_dialog_add_button (GTK_DIALOG (dlg),
				       GTK_STOCK_CANCEL,
				       GTK_RESPONSE_CLOSE);
		gtk_dialog_add_button (GTK_DIALOG (dlg),
				       GTK_STOCK_NEW,
				       GTK_RESPONSE_OK);
	} else {
		gtk_dialog_add_button (GTK_DIALOG (dlg),
				       GTK_STOCK_CLOSE,
				       GTK_RESPONSE_CLOSE);
	}
	gtk_dialog_set_default_response (GTK_DIALOG (dlg),
					 GTK_RESPONSE_CLOSE);

	dlg->priv->property_size_group = gtk_size_group_new (GTK_SIZE_GROUP_HORIZONTAL);
	dlg->priv->criteria_size_group = gtk_size_group_new (GTK_SIZE_GROUP_HORIZONTAL);
	dlg->priv->entry_size_group = gtk_size_group_new (GTK_SIZE_GROUP_HORIZONTAL);
	dlg->priv->button_size_group = gtk_size_group_new (GTK_SIZE_GROUP_HORIZONTAL);

	if (dlg->priv->creating)
		gtk_window_set_title (GTK_WINDOW (dlg), _("Create Automatic Playlist"));
	else
		gtk_window_set_title (GTK_WINDOW (dlg), _("Edit Automatic Playlist"));

	gtk_container_set_border_width (GTK_CONTAINER (dlg), 5);
	gtk_box_set_spacing (GTK_BOX (GTK_DIALOG (dlg)->vbox), 2);
	gtk_dialog_set_has_separator (GTK_DIALOG (dlg), FALSE);

	xml = rb_glade_xml_new ("create-playlist.glade",
				"main_vbox",
				dlg);
	
	dlg->priv->disjunction_check = GTK_WIDGET (glade_xml_get_widget (xml, "disjunctionCheck"));
	dlg->priv->limit_check = GTK_WIDGET (glade_xml_get_widget (xml, "limitCheck"));
	dlg->priv->limit_entry = GTK_WIDGET (glade_xml_get_widget (xml, "limitEntry"));
	dlg->priv->limit_option = GTK_WIDGET (glade_xml_get_widget (xml, "limitOption"));
	dlg->priv->sort_reverse_check = GTK_WIDGET (glade_xml_get_widget (xml, "descendingCheck"));
	dlg->priv->sort_option = GTK_WIDGET (glade_xml_get_widget (xml, "sortOrderMenu"));
	dlg->priv->addbutton = GTK_WIDGET (glade_xml_get_widget (xml, "addButton"));

	g_signal_connect_object (G_OBJECT (dlg->priv->limit_check), "toggled", G_CALLBACK (limit_toggled_cb),
				 dlg, 0);
	gtk_widget_set_sensitive (dlg->priv->limit_entry, FALSE);
	gtk_widget_set_sensitive (dlg->priv->limit_option, FALSE);
	gtk_size_group_add_widget (dlg->priv->button_size_group, dlg->priv->addbutton);
	g_signal_connect_object (G_OBJECT (dlg->priv->addbutton), "clicked", G_CALLBACK (add_button_click_cb),
				 dlg, 0);

	dlg->priv->vbox = GTK_BOX (glade_xml_get_widget (xml, "sub_vbox"));
	
	append_row(dlg);
	
	setup_sort_option_menu (dlg->priv->sort_option, sort_options,
						    G_N_ELEMENTS (sort_options),
						    FALSE);

	mainbox = glade_xml_get_widget (xml, "main_vbox");
	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dlg)->vbox), mainbox, FALSE, FALSE, 0);
	gtk_widget_show_all (GTK_WIDGET (dlg));

	return G_OBJECT (dlg);
}

static void
rb_query_creator_init (RBQueryCreator *dlg)
{
	dlg->priv = g_new0 (RBQueryCreatorPrivate, 1);
}

static void
rb_query_creator_finalize (GObject *object)
{
	RBQueryCreator *dlg;

	g_return_if_fail (object != NULL);
	g_return_if_fail (RB_IS_QUERY_CREATOR (object));

	dlg = RB_QUERY_CREATOR (object);

	g_return_if_fail (dlg->priv != NULL);

	g_object_unref (G_OBJECT (dlg->priv->property_size_group));
	g_object_unref (G_OBJECT (dlg->priv->criteria_size_group));
	g_object_unref (G_OBJECT (dlg->priv->entry_size_group));
	g_object_unref (G_OBJECT (dlg->priv->button_size_group));

	g_list_free (dlg->priv->rows);
	
	g_free (dlg->priv);

	G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
rb_query_creator_set_property (GObject *object,
			       guint prop_id,
			       const GValue *value,
			       GParamSpec *pspec)
{
	RBQueryCreator *dlg = RB_QUERY_CREATOR (object);

	switch (prop_id)
	{
	case PROP_DB:
		dlg->priv->db = g_value_get_object (value);
		break;
	case PROP_CREATING:
		dlg->priv->creating = g_value_get_boolean (value);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void 
rb_query_creator_get_property (GObject *object,
			      guint prop_id,
			      GValue *value,
			      GParamSpec *pspec)
{
	RBQueryCreator *dlg = RB_QUERY_CREATOR (object);

	switch (prop_id)
	{
	case PROP_DB:
		g_value_set_object (value, dlg->priv->db);
		break;
	case PROP_CREATING:
		g_value_set_boolean (value, dlg->priv->creating);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}


GtkWidget *
rb_query_creator_new (RhythmDB *db)
{
	return g_object_new (RB_TYPE_QUERY_CREATOR, "db", db, NULL);
}

static gboolean
rb_query_creator_set_query (RBQueryCreator *creator, GPtrArray *query)
{
	int i;
	GList *rows;
	gboolean disjunction = FALSE;
	RhythmDBQueryData *qdata;
	GPtrArray *subquery;
	
	g_return_val_if_fail(query->len == 2, FALSE);

	qdata = g_ptr_array_index (query, 1);
	g_return_val_if_fail(qdata->type == RHYTHMDB_QUERY_SUBQUERY, FALSE);

	subquery = qdata->subquery;

	if (subquery->len >0)
		for (i = 0; i < subquery->len - 1; i++) {
			RhythmDBQueryData *data = g_ptr_array_index (subquery, i);
			if (data->type != RHYTHMDB_QUERY_DISJUNCTION)
				append_row (creator);
		}
	
	rows = creator->priv->rows;

	for (i = 0; i < subquery->len; i++) {
		RhythmDBQueryData *data = g_ptr_array_index (subquery, i);
		GtkOptionMenu *propmenu;
		GtkOptionMenu *criteria_menu;
		gint i;

		if (data->type == RHYTHMDB_QUERY_DISJUNCTION) {
			disjunction = TRUE;
			continue;
		}

		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (creator->priv->disjunction_check),
				      disjunction);


		propmenu = GTK_OPTION_MENU (get_box_widget_at_pos (GTK_BOX (rows->data), 0));
		
		
		i = get_property_options_index(data->propid);
		if (i >= 0)
			gtk_option_menu_set_history (propmenu, i);
		else {
			rb_debug("Unable to handle property %d", data->propid);
			g_return_val_if_reached(FALSE);
		}

		criteria_menu = GTK_OPTION_MENU (get_box_widget_at_pos (GTK_BOX (rows->data),
									1));

		if (!select_criteria_from_value (creator,
					    GTK_WIDGET (criteria_menu), data->propid,
					    data->type))
			g_return_val_if_reached(FALSE);

		switch (property_options[i].entry_type) {
		case RHYTHM_QUERY_ENTRY_TEXT:
			{
				GtkEntry *text = GTK_ENTRY (get_box_widget_at_pos (GTK_BOX (rows->data), 2));
  			
				gtk_entry_set_text (text, g_value_get_string (data->val));
			}
			break;
			    
		case RHYTHM_QUERY_ENTRY_RATING:
			{
				RBRating *rating = RB_RATING (get_box_widget_at_pos (GTK_BOX (rows->data), 2));
				g_object_set (G_OBJECT (rating), "score",
					      g_value_get_double (data->val), NULL);
			}
			break;

		case RHYTHM_QUERY_ENTRY_POSITIVE_INTEGER:
			{
				GtkSpinButton *spinner = GTK_SPIN_BUTTON (get_box_widget_at_pos (GTK_BOX (rows->data), 2));

				gtk_spin_button_set_value (spinner, g_value_get_ulong (data->val));
			}
			break;
			
		case RHYTHM_QUERY_ENTRY_DURATION:
			{
				GtkBox *durationBox = GTK_BOX (get_box_widget_at_pos (GTK_BOX (rows->data), 2));
				
				GtkSpinButton *minutesSpinner = GTK_SPIN_BUTTON (get_box_widget_at_pos (durationBox, 0));
				GtkSpinButton *secondsSpinner = GTK_SPIN_BUTTON (get_box_widget_at_pos (durationBox, 2));
				
				gtk_spin_button_set_value (minutesSpinner, g_value_get_ulong (data->val) / 60);
				gtk_spin_button_set_value (secondsSpinner, g_value_get_ulong (data->val) % 60);
			}
			break;
		
		case RHYTHM_QUERY_ENTRY_RELATIVE_TIME:
			{
				GtkBox *box = GTK_BOX (get_box_widget_at_pos (GTK_BOX (rows->data), 2));
				
				GtkSpinButton *timeSpin = GTK_SPIN_BUTTON (get_box_widget_at_pos (box, 0));
				GtkOptionMenu *unitMenu = GTK_OPTION_MENU (get_box_widget_at_pos (box, 1));
				
				gulong time = g_value_get_ulong (data->val);
				gulong unit = 0;
				int i;
				
				/* determine the best units to use for the given value */
				for (i = 0; i < G_N_ELEMENTS(time_unit_options); i++) {
					/* find out if the time is an even multiple of the unit */
					if (time % time_unit_options[i].timeMultiplier == 0)
						unit = i;
				}
				
				/* set the time value and unit*/
				gtk_option_menu_set_history(unitMenu, unit);
				gtk_spin_button_set_value(timeSpin, time / time_unit_options[unit].timeMultiplier);
			}
			break;

		default:
			g_return_val_if_reached(FALSE);

		}
		
		rows = rows->next;
	}	
	return TRUE;
}

static gboolean
rb_query_creator_set_limits (RBQueryCreator *creator, int limit_count, int limit_size)
{
	guint limit_value = 0;

	if (limit_count) {
		gtk_option_menu_set_history (GTK_OPTION_MENU (creator->priv->limit_option), 0);
		limit_value = limit_count;
	} else if (limit_size > 0 && (limit_size % 1000 != 0)) {
		gtk_option_menu_set_history (GTK_OPTION_MENU (creator->priv->limit_option), 1);
		limit_value = limit_size;
	} else if (limit_size > 0) {
		gtk_option_menu_set_history (GTK_OPTION_MENU (creator->priv->limit_option), 2);
		limit_value = limit_size / 1000;
	}
	
	gtk_spin_button_set_value (GTK_SPIN_BUTTON (creator->priv->limit_entry), limit_value);
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (creator->priv->limit_check), limit_value != 0);
	
	return TRUE;
}

static gboolean
rb_query_creator_set_sorting (RBQueryCreator *creator, const char *sort_column, gint sort_direction)
{
	int i;

	if (!sort_column || ! *sort_column) {
		g_warning("No playlist sorting order");

		sort_column = DEFAULT_SORTING_COLUMN;
		sort_direction = DEFAULT_SORTING_ORDER;
	}

	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (creator->priv->sort_reverse_check), (sort_direction == GTK_SORT_DESCENDING));

	for (i = 0; i < G_N_ELEMENTS (sort_options); i++)
		if (strcmp (sort_options[i].sort_order, sort_column) == 0)
			break;

	/* check that it is a valid sort option */
	g_return_val_if_fail(i < G_N_ELEMENTS (sort_options), FALSE);
	

	gtk_option_menu_set_history (GTK_OPTION_MENU (creator->priv->sort_option), i);
	return TRUE;
}

GtkWidget *
rb_query_creator_new_from_query (RhythmDB *db, GPtrArray *query,
				 int limit_count, int limit_size,
				 const char *sort_column, gint sort_direction)
{
	RBQueryCreator *creator = g_object_new (RB_TYPE_QUERY_CREATOR, "db", db,
						"creating", FALSE, NULL);
	if (!creator)
		return NULL;

	if (! rb_query_creator_set_query (creator, query)
	 || ! rb_query_creator_set_limits (creator, limit_count, limit_size)
	 || ! rb_query_creator_set_sorting (creator, sort_column, sort_direction)) {
	   
	   /* the query creator could not be constructed. Show a warning and abort */
	   	GtkWidget *error_dialog;
	   	
		gtk_widget_destroy (GTK_WIDGET (creator));
	   
		error_dialog = gtk_message_dialog_new (
					NULL,
					0,
					GTK_MESSAGE_ERROR,
					GTK_BUTTONS_CLOSE,
					_("This automatic playlist cannot be edited with this version of Rhythmbox"));
		gtk_dialog_run (GTK_DIALOG (error_dialog));
		gtk_widget_destroy (error_dialog);	
		return NULL;
	}
	
	return GTK_WIDGET (creator);
}

static GtkWidget *
get_box_widget_at_pos (GtkBox *box, guint pos)
{
	GtkWidget *ret = NULL;
	GList *children = gtk_container_get_children (GTK_CONTAINER (box));
	GList *tem;
	for (tem = children; tem; tem = tem->next) {
		GValue thispos = { 0, };
		g_value_init (&thispos, G_TYPE_INT);
		gtk_container_child_get_property (GTK_CONTAINER (box),
						  GTK_WIDGET (tem->data),
						  "position", &thispos);
		if (g_value_get_int (&thispos) == pos) {
			ret = tem->data;
			break;
		}
	}
	g_list_free (children);
	return GTK_WIDGET (ret);
}

static gint
get_property_options_index(RhythmDBPropType prop) {
	int i;

	for (i = 0; i < G_N_ELEMENTS (property_options); i++)
		if (prop == property_options[i].strict_val || prop == property_options[i].fuzzy_val)
			return i;

	g_return_val_if_reached(-1);
}


static const RBQueryCreatorCriteriaOption *
get_criteria_for_property (RBQueryCreator *creator, RhythmDBPropType prop, guint *length)
{
	int i = get_property_options_index (prop);
	
	if (i < 0)
		return NULL;
	
	if (length)
		*length = property_options[i].criteria_options_length;

	return property_options[i].criteria_options;
}

static void
set_rating_score (RBRating *rating, gdouble score)
{
	g_object_set (G_OBJECT (rating), "score", score, NULL);
}	

static GtkWidget *
get_entry_for_property (RBQueryCreator *creator, const RBQueryCreatorPropertyOption *prop_option,
			gboolean *constrain)
{
	GtkWidget *widget = NULL;
	
	*constrain = TRUE;

	switch (prop_option->entry_type) {
	case RHYTHM_QUERY_ENTRY_TEXT:
		widget = gtk_entry_new ();
		break;
		
	case RHYTHM_QUERY_ENTRY_RATING:
		widget = GTK_WIDGET(rb_rating_new ());
		g_signal_connect_object (G_OBJECT (widget), "rated",
					 G_CALLBACK (set_rating_score), NULL, 0);
		*constrain = FALSE;
		break;
		
	case RHYTHM_QUERY_ENTRY_POSITIVE_INTEGER:
		widget = gtk_spin_button_new_with_range (0.0, 10000.0, 1.0);
		break;
	
	case RHYTHM_QUERY_ENTRY_DURATION:
		{
			GtkWidget *minutesSpin;
			GtkWidget *minutesLabel;
			GtkWidget *secondsSpin;

			widget = gtk_hbox_new(FALSE, 6);
			
			/* the criteria for Duration is set out like the following [ 2] : [30] */
			
			minutesSpin = gtk_spin_button_new_with_range (0.0, 10000.0, 1.0);
			gtk_box_pack_start_defaults (GTK_BOX(widget), minutesSpin);
			
			minutesLabel = gtk_label_new ( _(":") );
			gtk_box_pack_start (GTK_BOX(widget), minutesLabel, FALSE, FALSE, 0);
			
			secondsSpin = gtk_spin_button_new_with_range (0.0, 59.0, 1.0);
			gtk_box_pack_start_defaults (GTK_BOX(widget), secondsSpin);
			
			gtk_widget_show_all (widget);
		}
		break;
	
 	case RHYTHM_QUERY_ENTRY_RELATIVE_TIME:
		{
			GtkWidget *timeSpin;
			GtkWidget *timeOption;

			widget = gtk_hbox_new(FALSE, 6);
			
			timeSpin = gtk_spin_button_new_with_range (1.0, 60.0, 1.0);
			gtk_box_pack_start_defaults (GTK_BOX(widget), timeSpin);
			
			timeOption = create_time_unit_option_menu(time_unit_options, G_N_ELEMENTS(time_unit_options), FALSE);
			gtk_option_menu_set_history(GTK_OPTION_MENU (timeOption), time_unit_options_default);
			gtk_box_pack_start_defaults (GTK_BOX (widget), timeOption);
				
			gtk_widget_show_all (widget);
		}
		break;
		
	default:
		g_return_val_if_reached(NULL);
	}

	return widget;
}

static int
extract_option_menu_index (GtkOptionMenu *menu)
{
	GtkWidget *active_item = option_menu_get_active_child (GTK_WIDGET (menu));
	return GPOINTER_TO_INT (g_object_get_data (G_OBJECT (active_item), "rb-query-creator-value"));
}

GPtrArray *
rb_query_creator_get_query (RBQueryCreator *dlg)
{
	GPtrArray *query;
	GPtrArray *sub_query;
	GList *rows, *row;
	gboolean disjunction;

	query = rhythmdb_query_parse (dlg->priv->db,
				      RHYTHMDB_QUERY_PROP_EQUALS,
				      RHYTHMDB_PROP_TYPE,
				      RHYTHMDB_ENTRY_TYPE_SONG,
				      RHYTHMDB_QUERY_END);

	disjunction = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (dlg->priv->disjunction_check));
	
	sub_query = g_ptr_array_new ();
	
	rows = dlg->priv->rows;
	for (row = rows; row; row = row->next) {
		GtkOptionMenu *propmenu = GTK_OPTION_MENU (get_box_widget_at_pos (GTK_BOX (row->data),
										  0));
		GtkOptionMenu *criteria_menu = GTK_OPTION_MENU (get_box_widget_at_pos (GTK_BOX (row->data),
										       1));

		const RBQueryCreatorPropertyOption *prop_option = &property_options[extract_option_menu_index (propmenu)];
		const RBQueryCreatorCriteriaOption *criteria_options = get_criteria_for_property (dlg, prop_option->strict_val, NULL);
		const RBQueryCreatorCriteriaOption *criteria_option;
		RhythmDBQueryType criteria;
		RhythmDBPropType prop;

		g_return_val_if_fail(criteria_options != NULL, NULL);

		criteria_option = &criteria_options[extract_option_menu_index (criteria_menu)];
		criteria = criteria_option->val;
		prop = criteria_option->strict ? prop_option->strict_val : prop_option->fuzzy_val;

		switch (prop_option->entry_type) {
		case RHYTHM_QUERY_ENTRY_TEXT:
			{
				GtkEntry *text = GTK_ENTRY (get_box_widget_at_pos (GTK_BOX (row->data), 2));
				const char *data = gtk_entry_get_text (GTK_ENTRY (text));
				char *folded_data = criteria_option->strict ? g_strdup (data) : g_utf8_casefold (data, -1);
				
				/* ignore if it is enpty */
				if (folded_data && *folded_data) {
					if (disjunction && row->next)
						rhythmdb_query_append (dlg->priv->db,
								       sub_query,
								       criteria,
								       prop,
								       folded_data,
								       RHYTHMDB_QUERY_DISJUNCTION,
								       RHYTHMDB_QUERY_END);
					else
						rhythmdb_query_append (dlg->priv->db,
								       sub_query,
								       criteria,
								       prop,
								       folded_data,
								       RHYTHMDB_QUERY_END);
				}
				g_free (folded_data);
			}
			break;

		case RHYTHM_QUERY_ENTRY_RATING:
			{
				double rating;

				g_object_get (G_OBJECT (get_box_widget_at_pos (GTK_BOX (row->data), 2)),
					      "score", &rating, NULL);
				if (disjunction && row->next)
					rhythmdb_query_append (dlg->priv->db,
							       sub_query,
							       criteria,
							       prop,
							       rating,
							       RHYTHMDB_QUERY_DISJUNCTION,
							       RHYTHMDB_QUERY_END);
				else
					rhythmdb_query_append (dlg->priv->db,
							       sub_query,
							       criteria,
							       prop,
							       rating,
							       RHYTHMDB_QUERY_END);
			}
			break;

		case RHYTHM_QUERY_ENTRY_POSITIVE_INTEGER:
			{
				GtkSpinButton *spinner = GTK_SPIN_BUTTON (get_box_widget_at_pos (GTK_BOX (row->data), 2));
				gulong value = gtk_spin_button_get_value_as_int(spinner);

				if (disjunction && row->next)
					rhythmdb_query_append (dlg->priv->db,
							       sub_query,
							       criteria,
							       prop,
							       value,
							       RHYTHMDB_QUERY_DISJUNCTION,
							       RHYTHMDB_QUERY_END);
				else
					rhythmdb_query_append (dlg->priv->db,
							       sub_query,
							       criteria,
							       prop,
							       value,
							       RHYTHMDB_QUERY_END);
			}
			break;
		
		case RHYTHM_QUERY_ENTRY_DURATION:
			{
				GtkBox *durationBox = GTK_BOX (get_box_widget_at_pos (GTK_BOX (row->data), 2));
				
				GtkSpinButton *minutesSpinner = GTK_SPIN_BUTTON (get_box_widget_at_pos (durationBox, 0));
				GtkSpinButton *secondsSpinner = GTK_SPIN_BUTTON (get_box_widget_at_pos (durationBox, 2));
				
				gulong value = gtk_spin_button_get_value_as_int(minutesSpinner) * 60
					     + gtk_spin_button_get_value_as_int(secondsSpinner);

				if (disjunction && row->next)
					rhythmdb_query_append (dlg->priv->db,
							       sub_query,
							       criteria,
							       prop,
							       value,
							       RHYTHMDB_QUERY_DISJUNCTION,
							       RHYTHMDB_QUERY_END);
				else
					rhythmdb_query_append (dlg->priv->db,
							       sub_query,
							       criteria,
							       prop,
							       value,
							       RHYTHMDB_QUERY_END);
			}
			break;
		
		case RHYTHM_QUERY_ENTRY_RELATIVE_TIME:
			{
				GtkBox *box = GTK_BOX (get_box_widget_at_pos (GTK_BOX (row->data), 2));
				
				GtkSpinButton *timeSpin = GTK_SPIN_BUTTON (get_box_widget_at_pos (box, 0));
				GtkOptionMenu *unitMenu = GTK_OPTION_MENU (get_box_widget_at_pos (box, 1));
				
				gulong timeMultiplier = time_unit_options[gtk_option_menu_get_history(unitMenu)].timeMultiplier;
				gulong value = gtk_spin_button_get_value_as_int(timeSpin) * timeMultiplier;

				if (disjunction && row->next)
					rhythmdb_query_append (dlg->priv->db,
							       sub_query,
							       criteria,
							       prop,
							       value,
							       RHYTHMDB_QUERY_DISJUNCTION,
							       RHYTHMDB_QUERY_END);
				else
					rhythmdb_query_append (dlg->priv->db,
							       sub_query,
							       criteria,
							       prop,
							       value,
							       RHYTHMDB_QUERY_END);
			}
			break;

		default:
			g_return_val_if_reached (NULL);
		}
	}
	rhythmdb_query_append (dlg->priv->db,
			       query,
			       RHYTHMDB_QUERY_SUBQUERY,
			       sub_query,
			       RHYTHMDB_QUERY_END);
	return query;
}

void
rb_query_creator_get_limit (RBQueryCreator *dlg, RBQueryCreatorLimitType *type,
			    guint *limit)
{
	guint limitpos;
	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (dlg->priv->limit_check)))
		*limit = atoi (gtk_entry_get_text (GTK_ENTRY (dlg->priv->limit_entry)));
	else
		*limit = 0;
	limitpos = gtk_option_menu_get_history (GTK_OPTION_MENU (dlg->priv->limit_option));
	switch (limitpos)
	{
	case 0:
		*type = RB_QUERY_CREATOR_LIMIT_COUNT;
		break;
	case 1:
		*type = RB_QUERY_CREATOR_LIMIT_MB;
		break;
	case 2:
		*type = RB_QUERY_CREATOR_LIMIT_MB;
		*limit *= 1000;
		break;
	default:
		g_assert_not_reached ();
	}
}

void
rb_query_creator_get_sort_order (RBQueryCreator *dlg, const char **sort_column, gint *sort_direction) {
	int i;

	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (dlg->priv->sort_reverse_check)))
		*sort_direction = GTK_SORT_DESCENDING;
	else
		*sort_direction = GTK_SORT_ASCENDING;

	i = gtk_option_menu_get_history (GTK_OPTION_MENU (dlg->priv->sort_option));
	*sort_column = sort_options[i].sort_order;
}

static void
limit_toggled_cb (GtkWidget *limit, RBQueryCreator *dialog)
{
	gboolean active = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (limit));

	gtk_widget_set_sensitive (dialog->priv->limit_entry, active);
	gtk_widget_set_sensitive (dialog->priv->limit_option, active);
}

static GtkWidget *
lookup_row_by_widget (RBQueryCreator *dialog, GtkWidget *widget)
{
	GList *rows = dialog->priv->rows;
	GList *row;
	GtkWidget *ret = NULL;
	guint i;

	for (row = rows, i = 0; row; row = row->next, i++) {
		GList *columns = gtk_container_get_children (GTK_CONTAINER (row->data));
		gboolean found = g_list_find (columns, widget) != NULL;
		g_list_free (columns);
		if (found) {
			ret = row->data;
			break;
		}
	}
	return ret;
}

static void
remove_button_click_cb (GtkWidget *button, RBQueryCreator *dialog)
{
	GtkWidget *row;

	row = lookup_row_by_widget (dialog, button);
	g_assert (row);
	gtk_container_remove (GTK_CONTAINER (dialog->priv->vbox),
			      GTK_WIDGET (row));
	dialog->priv->rows = g_list_remove (dialog->priv->rows, row);
}

static void
add_button_click_cb (GtkWidget *button, RBQueryCreator *creator)
{
	append_row (creator);
}

static GtkWidget *
append_row (RBQueryCreator *dialog)
{
	GtkWidget *option;
	GtkWidget *criteria;
	GtkWidget *entry;
	GtkWidget *remove_button;
	GtkBox *hbox;
	GList *rows;
	guint len;
	gboolean constrain;

	rows = dialog->priv->rows;
	len = g_list_length (rows);

	hbox = GTK_BOX (gtk_hbox_new (FALSE, 5));
	gtk_box_pack_start_defaults (GTK_BOX (dialog->priv->vbox), GTK_WIDGET (hbox));
	dialog->priv->rows = g_list_prepend (dialog->priv->rows, hbox);
	gtk_box_reorder_child (dialog->priv->vbox, GTK_WIDGET (hbox), -1);

	/* Set up the defaunt property/criteria/entry widgets. */
	option = create_property_option_menu (dialog, property_options,
					      G_N_ELEMENTS (property_options),
					      FALSE);
	gtk_size_group_add_widget (dialog->priv->property_size_group, option);
	gtk_box_pack_start_defaults (hbox, GTK_WIDGET (option));
	gtk_option_menu_set_history (GTK_OPTION_MENU (option), 0);
	criteria = create_criteria_option_menu (property_options[0].criteria_options,
						property_options[0].criteria_options_length,
						FALSE);
	gtk_size_group_add_widget (dialog->priv->criteria_size_group, criteria);
	gtk_box_pack_start_defaults (hbox, GTK_WIDGET (criteria));

	entry = get_entry_for_property (dialog, &property_options[0],
					&constrain);
	g_return_val_if_fail(entry != NULL, NULL);
	gtk_widget_show (entry);
	if (constrain)
		gtk_size_group_add_widget (dialog->priv->entry_size_group, entry);
	gtk_box_pack_start_defaults (GTK_BOX (hbox), GTK_WIDGET (entry));

	/* Let users remove the default row */
	remove_button = gtk_button_new_from_stock (GTK_STOCK_REMOVE);
	g_signal_connect_object (G_OBJECT (remove_button), "clicked", G_CALLBACK (remove_button_click_cb),
				 dialog, 0);
	gtk_size_group_add_widget (dialog->priv->button_size_group, remove_button);
	gtk_box_pack_start_defaults (hbox, GTK_WIDGET (remove_button));

	gtk_widget_show_all (GTK_WIDGET (dialog->priv->vbox));
	return GTK_WIDGET (hbox);
}

static gboolean
select_criteria_from_value (RBQueryCreator *creator,
			    GtkWidget *option_menu,
			    RhythmDBPropType prop,
			    RhythmDBQueryType qtype)
{
	guint i;
	const RBQueryCreatorCriteriaOption *options;
	guint length;

	options = get_criteria_for_property (creator, prop, &length);
	
	if (options == NULL)
		return FALSE;
			
	for (i = 0; i < length; i++) {
		if (qtype == options[i].val) {
			gtk_option_menu_set_history (GTK_OPTION_MENU (option_menu), i);
			return TRUE;
		}
	}
	return FALSE;
}

static void
property_option_menu_changed (GtkOptionMenu *propmenu,
			      RBQueryCreator *creator)
{
	GtkWidget *row;
	GtkWidget *criteria;
	GtkWidget *entry;
	const RBQueryCreatorPropertyOption *prop_option;
	const RBQueryCreatorCriteriaOption *criteria_options;
	guint length;
	gboolean constrain;

	row = lookup_row_by_widget (creator, GTK_WIDGET (propmenu));
	
	

	criteria = get_box_widget_at_pos (GTK_BOX (row), 1);
	gtk_container_remove (GTK_CONTAINER (row), criteria);

	prop_option = &property_options[extract_option_menu_index (propmenu)];

	criteria_options = get_criteria_for_property (creator, prop_option->strict_val, &length);
	g_return_if_fail(criteria_options != NULL);

	criteria = create_criteria_option_menu (criteria_options, length,
						FALSE);
	gtk_widget_show (criteria);
	gtk_size_group_add_widget (creator->priv->criteria_size_group, criteria);
	gtk_box_pack_start_defaults (GTK_BOX (row), GTK_WIDGET (criteria));
	gtk_box_reorder_child (GTK_BOX (row), criteria, 1);

	entry = get_box_widget_at_pos (GTK_BOX (row), 2);
	gtk_container_remove (GTK_CONTAINER (row), entry);
	entry = get_entry_for_property (creator, prop_option,
					&constrain);
	g_return_if_fail(entry != NULL);
	gtk_widget_show (entry);
	if (constrain)
		gtk_size_group_add_widget (creator->priv->entry_size_group, entry);
	gtk_box_pack_start_defaults (GTK_BOX (row), GTK_WIDGET (entry));
	gtk_box_reorder_child (GTK_BOX (row), entry, 2);
}

/* Stolen from jamboree and hacked up */
static void
setup_property_option_menu (RBQueryCreator *creator,
			    GtkWidget *option_menu,
			    const RBQueryCreatorPropertyOption *options,
			    int length, gboolean activate_first)
{
	GtkWidget *menu;
	GtkWidget *menu_item;
	int i;

	menu = gtk_menu_new ();
	gtk_widget_show (menu);
    
	for (i = 0; i < length; i++) {
		menu_item = gtk_menu_item_new_with_label (_(options[i].name));
		g_object_set_data (G_OBJECT (menu_item),
				   "rb-query-creator-value", GINT_TO_POINTER (i));
		gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
		gtk_widget_show (menu_item);
	}

	if (activate_first)
		gtk_option_menu_set_history (GTK_OPTION_MENU (option_menu), 0);

	gtk_option_menu_set_menu (GTK_OPTION_MENU (option_menu), menu);

	g_signal_connect_object (G_OBJECT (option_menu), "changed",
				 G_CALLBACK (property_option_menu_changed), creator, 0);
}

static void
setup_criteria_option_menu (GtkWidget *option_menu,
			    const RBQueryCreatorCriteriaOption *options,
			    int length, gboolean activate_first)
{
	GtkWidget *menu;
	GtkWidget *menu_item;
	int i;

	menu = gtk_menu_new ();
	gtk_widget_show (menu);
    
	for (i = 0; i < length; i++) {
		menu_item = gtk_menu_item_new_with_label (_(options[i].name));
		g_object_set_data (G_OBJECT (menu_item),
				   "rb-query-creator-value", GINT_TO_POINTER (i));
		gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
		gtk_widget_show (menu_item);
	}

	if (activate_first)
		gtk_option_menu_set_history (GTK_OPTION_MENU (option_menu), 0);

	gtk_option_menu_set_menu (GTK_OPTION_MENU (option_menu), menu);
}

static void
setup_time_unit_option_menu (GtkWidget *option_menu,
			    const RBQueryCreatorTimeUnitOption *options,
			    int length, gboolean activate_first)
{
	GtkWidget *menu;
	GtkWidget *menu_item;
	int i;

	menu = gtk_menu_new ();
	gtk_widget_show (menu);
    
	for (i = 0; i < length; i++) {
		menu_item = gtk_menu_item_new_with_label (_(options[i].name));
		g_object_set_data (G_OBJECT (menu_item),
				   "rb-query-creator-value", GINT_TO_POINTER (i));
		gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
		gtk_widget_show (menu_item);
	}

	if (activate_first)
		gtk_option_menu_set_history (GTK_OPTION_MENU (option_menu), 0);

	gtk_option_menu_set_menu (GTK_OPTION_MENU (option_menu), menu);
}

static void
setup_sort_option_menu (GtkWidget *option_menu,
			    const RBQueryCreatorSortOption *options,
			    int length, gboolean activate_first)
{
	GtkWidget *menu;
	GtkWidget *menu_item;
	int i;

	menu = gtk_menu_new ();
	gtk_widget_show (menu);
    
	for (i = 0; i < length; i++) {
		menu_item = gtk_menu_item_new_with_label (_(options[i].name));
		g_object_set_data (G_OBJECT (menu_item),
				   "rb-query-creator-value", GINT_TO_POINTER (i));
		gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
		gtk_widget_show (menu_item);
	}

	if (activate_first)
		gtk_option_menu_set_history (GTK_OPTION_MENU (option_menu), 0);
		
	gtk_option_menu_set_menu (GTK_OPTION_MENU (option_menu), menu);
}

static GtkWidget*
create_property_option_menu (RBQueryCreator *creator,
			     const RBQueryCreatorPropertyOption *options,
			     int length, gboolean activate_first)
{
	GtkWidget *menu;
	GtkWidget *option_menu;
  
	menu = gtk_menu_new ();
	gtk_widget_show (menu);
    
	option_menu = gtk_option_menu_new ();

	setup_property_option_menu (creator, option_menu, options, length, activate_first);
  
	return option_menu;
}

static GtkWidget*
create_criteria_option_menu (const RBQueryCreatorCriteriaOption *options,
			     int length, gboolean activate_first)
{
	GtkWidget *menu;
	GtkWidget *option_menu;
  
	menu = gtk_menu_new ();
	gtk_widget_show (menu);
    
	option_menu = gtk_option_menu_new ();

	setup_criteria_option_menu (option_menu, options, length, activate_first);
  
	return option_menu;
}

static GtkWidget*
create_time_unit_option_menu (const RBQueryCreatorTimeUnitOption *options,
			     int length, gboolean activate_first)
{
	GtkWidget *menu;
	GtkWidget *option_menu;
  
	menu = gtk_menu_new ();
	gtk_widget_show (menu);
    
	option_menu = gtk_option_menu_new ();

	setup_time_unit_option_menu (option_menu, options, length, activate_first);
  
	return option_menu;
}

static GtkWidget*
option_menu_get_active_child (GtkWidget *option_menu)
{
	GtkWidget *menu;
	GList *children;
	int pos;
	GtkWidget *child;
  
	menu = gtk_option_menu_get_menu (GTK_OPTION_MENU (option_menu));
	children = gtk_container_get_children (GTK_CONTAINER (menu));

	pos = gtk_option_menu_get_history (GTK_OPTION_MENU (option_menu));
	child = g_list_nth (children, pos)->data;
	g_assert (child != NULL);
	g_assert (GTK_IS_WIDGET (child));
  
	return child;
}

